import asyncio
import copy
import dataclasses
import difflib
from asyncio import CancelledError
from collections import defaultdict
from datetime import datetime, timezone
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Dict,
    List,
    Optional,
    Text,
    Tuple,
    Union,
)
from urllib.parse import urlparse

import requests
import structlog
from tqdm import tqdm

import rasa.shared.utils.io
from rasa.core.channels import CollectingOutputChannel, UserMessage
from rasa.core.config.available_endpoints import AvailableEndpoints
from rasa.core.config.configuration import Configuration
from rasa.core.constants import (
    ACTIVE_FLOW_METADATA_KEY,
    PARENT_FLOW_ID_METADATA_KEY,
    PARENT_STEP_ID_METADATA_KEY,
    STEP_ID_METADATA_KEY,
)
from rasa.core.exceptions import AgentNotReady
from rasa.core.persistor import StorageType
from rasa.dialogue_understanding_test.du_test_case import DialogueUnderstandingTestCase
from rasa.e2e_test.constants import TEST_CASE_NAME, TEST_FILE_NAME
from rasa.e2e_test.e2e_config import create_llm_judge_config
from rasa.e2e_test.e2e_test_case import (
    KEY_STUB_CUSTOM_ACTIONS,
    ActualStepOutput,
    Fixture,
    Metadata,
    TestCase,
    TestStep,
)
from rasa.e2e_test.e2e_test_result import (
    NO_RESPONSE,
    NO_SLOT,
    TestFailure,
    TestResult,
)
from rasa.exceptions import ValidationError
from rasa.llm_fine_tuning.conversations import Conversation
from rasa.shared.agents.agent_setup import AgentsConnectionCleanup
from rasa.shared.constants import RASA_DEFAULT_FLOW_PATTERN_PREFIX
from rasa.shared.core.constants import MOCKED_DATETIME_SLOT
from rasa.shared.core.events import (
    ActionExecuted,
    BotUttered,
    Event,
    FlowCompleted,
    FlowStarted,
    SlotSet,
    UserUttered,
)
from rasa.shared.core.flows.flow_path import FlowPath, PathNode
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import RasaException
from rasa.shared.nlu.constants import COMMANDS
from rasa.telemetry import track_e2e_test_run
from rasa.utils.endpoints import EndpointConfig

if TYPE_CHECKING:
    from rasa.core.agent import Agent
    from rasa.shared.core.flows import FlowsList

structlogger = structlog.get_logger()

TEST_TURNS_TYPE = Dict[int, Union[TestStep, ActualStepOutput]]


class E2ETestRunner:
    def __init__(
        self,
        model_path: Optional[Text] = None,
        model_server: Optional[EndpointConfig] = None,
        remote_storage: Optional[StorageType] = None,
        endpoints: Optional[AvailableEndpoints] = None,
        sub_agents_path: Optional[Text] = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the E2E test suite runner.

        Args:
            model_path: Path to the model.
            model_server: Model server configuration.
            remote_storage: Remote storage to use for model retrieval.
            endpoints: Endpoints configuration.
            **kwargs: Additional arguments
        """
        import rasa.core.agent

        structlogger.info(
            "e2e_test_runner.init",
            event_info="Started running end-to-end testing.",
        )

        test_case_path = kwargs.get("test_case_path")
        self.llm_judge_config = create_llm_judge_config(test_case_path)

        are_custom_actions_stubbed = (
            endpoints
            and endpoints.action
            and endpoints.action.kwargs.get(KEY_STUB_CUSTOM_ACTIONS)
        )
        if endpoints and not are_custom_actions_stubbed:
            self._action_server_is_reachable(endpoints)

        sub_agents = Configuration.get_instance().available_agents

        async def load_agent_with_agents_connection_cleanup() -> "Agent":
            # Create a wrapper coroutine that handles graceful agents connection
            # cleanup, this prevents abrupt closure when asyncio.run finishes.
            async with AgentsConnectionCleanup():
                return await rasa.core.agent.load_agent(
                    model_path=model_path,
                    model_server=model_server,
                    remote_storage=remote_storage,
                    endpoints=endpoints,
                    sub_agents=sub_agents,
                )
            # Defensive return for mypy - never actually reached
            return

        self.agent = asyncio.run(load_agent_with_agents_connection_cleanup())

        if not self.agent.is_ready():
            raise AgentNotReady(
                "Agent needs to be prepared before usage. "
                "Please check that the agent was able to "
                "load the trained model."
            )

    async def run_prediction_loop(
        self,
        collector: CollectingOutputChannel,
        steps: List[TestStep],
        sender_id: Text,
        test_case_metadata: Optional[Metadata] = None,
        input_metadata: Optional[List[Metadata]] = None,
    ) -> TEST_TURNS_TYPE:
        """Runs dialogue prediction.

        Args:
            collector: Output channel.
            steps: List of steps to run.
            sender_id: The test case name with added timestamp suffix.
            test_case_metadata: Metadata of the test case.
            input_metadata: List of metadata.

        Returns:
            Test turns: {turn_sequence (int): TestStep or ActualStepOutput}.
        """
        turns: TEST_TURNS_TYPE = {}
        event_cursor = 0

        if not self.agent.processor:
            return turns

        tracker = await self.agent.processor.fetch_tracker_with_initial_session(
            sender_id, output_channel=collector
        )
        # turn -1 i used to contain events that happen during
        # the start of the session and before the first user message
        # TestStep is a placeholder just for the sake of having a turn
        # to specify the actor
        turns[-1], event_cursor = self.get_actual_step_output(
            tracker,
            TestStep(
                actor="bot",
                text=None,
            ),
            event_cursor,
        )

        # This variable tracks the position of the last user step.
        # We use it to determine from which turn onward we should check
        # for failures in case of the fail-fast logic.
        last_user_step_position = 0
        for position, step in enumerate(steps):
            if step.actor != "user":
                turns[position] = step

                # Check for failures after each bot step
                try:
                    self.fail_fast_check(
                        test_turns=turns,
                        last_user_step_position=last_user_step_position,
                    )
                except RasaException:
                    break

                continue
            elif not step.text:
                rasa.shared.utils.io.raise_warning(
                    f"The test case '{sender_id}' contains a `user` step in line "
                    f"{position + 1} without a text value. "
                    f"Skipping this step and proceeding to the next user step.",
                    UserWarning,
                )
                continue
            # Update to the current position because we have a valid user step
            last_user_step_position = position

            metadata = test_case_metadata.metadata if test_case_metadata else {}

            if input_metadata:
                step_metadata = self.filter_metadata_for_input(
                    step.metadata_name, input_metadata
                )
                step_metadata_dict = step_metadata.metadata if step_metadata else {}
                metadata = self.merge_metadata(
                    sender_id, step.text, metadata, step_metadata_dict
                )

            try:
                await self.agent.handle_message(
                    UserMessage(
                        step.text,
                        collector,
                        sender_id,
                        metadata=metadata,
                    )
                )
            except CancelledError:
                structlogger.error(
                    "e2e_test_runner.run_prediction_loop",
                    error=f"Message handling timed out for user message '{step.text}'.",
                    exc_info=True,
                )
            except Exception as exc:
                structlogger.error(
                    "e2e_test_runner.run_prediction_loop",
                    error=f"An exception occurred while handling "
                    f"user message '{step.text}'. Error: {exc}",
                )
            tracker = await self.agent.tracker_store.retrieve(sender_id)  # type: ignore[assignment]
            turns[position], event_cursor = self.get_actual_step_output(
                tracker, step, event_cursor
            )
        return turns

    def fail_fast_check(
        self, test_turns: TEST_TURNS_TYPE, last_user_step_position: int
    ) -> None:
        """Checks whether there are any test failures in 'test_turns'.

        Args:
            test_turns (TEST_TURNS_TYPE): The transcript of test cases and events.
            last_user_step_position (int): The start position in the test turns.

        Raises:
            RasaException if a failure is found.
        """
        failures_found = self.find_test_failures(
            test_turns=test_turns,
            test_case=None,
            last_user_step_position=last_user_step_position,
        )
        if failures_found:
            raise RasaException("Test failure found. Aborting early.")

    @staticmethod
    def merge_metadata(
        sender_id: Text,
        step_text: Text,
        test_case_metadata: Dict[Text, Text],
        step_metadata: Dict[Text, Text],
    ) -> Dict[Text, Text]:
        """Merges the test case and user step metadata.

        Args:
            sender_id: The test case name with added timestamp suffix.
            step_text: The user step text.
            test_case_metadata: The test case metadata dict.
            step_metadata: The user step metadata dict.

        Returns:
            A dictionary with the merged metadata.
        """
        if not test_case_metadata:
            return step_metadata
        if not step_metadata:
            return test_case_metadata

        keys_to_overwrite = []

        for key in step_metadata.keys():
            if key in test_case_metadata.keys():
                keys_to_overwrite.append(key)

        if keys_to_overwrite:
            test_case_name = sender_id.rsplit("_", 1)[0]
            structlogger.warning(
                "e2e_test_runner.merge_metadata",
                message=f"Metadata {keys_to_overwrite} exist in both the test case "
                f"'{test_case_name}' and the user step '{step_text}'. "
                "The user step metadata takes precedence and will "
                "override the test case metadata.",
            )

        merged_metadata = copy.deepcopy(test_case_metadata)
        merged_metadata.update(step_metadata)

        return merged_metadata

    @staticmethod
    def get_actual_step_output(
        tracker: DialogueStateTracker,
        test_step: TestStep,
        event_cursor: int,
    ) -> Tuple[ActualStepOutput, int]:
        """Returns the events that are generated from the current step.

        Args:
            tracker: The tracker for the current test case.
            test_step: The test step.
            event_cursor: The event cursor where the previous step left off.

        Returns:
        The events generated from the current step and the updated event
        cursor position.
        """
        session_events = list(tracker.events)[event_cursor:]
        if session_events:
            event_cursor += len(session_events)
            return (
                ActualStepOutput.from_test_step(
                    test_step,
                    [
                        event
                        for event in session_events
                        if isinstance(event, (BotUttered, UserUttered, SlotSet))
                    ],
                ),
                event_cursor,
            )
        else:
            structlogger.warning(
                "e2e_test_runner.get_actual_step_output",
                message=f"No events found for '{tracker.sender_id}' after processing "
                f"test step '{test_step.text}'.",
            )
            # if there are no events, we still want to return an
            # ActualStepOutput object with the test step as the
            # response to be able to generate a diff
            return (
                ActualStepOutput.from_test_step(
                    test_step,
                    [
                        UserUttered(text=test_step.text),
                        BotUttered(text=NO_RESPONSE),
                    ],
                ),
                event_cursor,
            )

    @classmethod
    def generate_test_result(
        cls,
        test_turns: TEST_TURNS_TYPE,
        test_case: TestCase,
    ) -> TestResult:
        """Generates test result.

        Args:
            test_turns: the turns that happened when running the test case or test step.
            test_case: the `TestCase` instance.

        Returns:
        Test result.
        """
        difference = []
        error_line = None
        test_failures = cls.find_test_failures(test_turns, test_case)
        if test_failures:
            first_failure = test_failures[0][0]
            difference = cls.human_readable_diff(test_turns, test_failures)
            error_line = first_failure.error_line if first_failure else None

        return TestResult(
            pass_status=len(test_failures) == 0,
            test_case=test_case,
            difference=difference,
            error_line=error_line,
        )

    def _get_additional_splitting_conditions(
        self,
        step: TestStep,
        input_metadata: List[Metadata],
        tracker: DialogueStateTracker,
        test_case: TestCase,
    ) -> Dict[str, Any]:
        """Returns additional splitting conditions for the user message."""
        additional_splitting_conditions: Dict[str, Any] = {"text": step.text}

        if not step.metadata_name:
            return additional_splitting_conditions

        step_metadata = self.filter_metadata_for_input(
            step.metadata_name, input_metadata
        )
        step_metadata_dict = step_metadata.metadata if step_metadata else {}

        test_case_metadata = self.filter_metadata_for_input(
            test_case.metadata_name, input_metadata
        )
        test_case_metadata_as_dict = (
            test_case_metadata.metadata if test_case_metadata else {}
        )

        metadata: Dict[str, Any] = self.merge_metadata(
            tracker.sender_id,
            step.text,
            test_case_metadata_as_dict,
            step_metadata_dict,
        )
        metadata["model_id"] = tracker.model_id
        metadata["assistant_id"] = tracker.assistant_id

        additional_splitting_conditions["metadata"] = metadata

        return additional_splitting_conditions

    @staticmethod
    def _get_current_user_turn_and_prior_events(
        tracker: DialogueStateTracker,
        additional_splitting_conditions: Dict[str, Any],
        step: TestStep,
    ) -> Tuple[List[Event], List[Event]]:
        """Returns the current user turn and prior events."""
        actual_events = tracker.events

        # this returns 2 lists, the first list contains the events until the user
        # message and the second list contains the events after the
        # user message, including the user message
        step_events = rasa.shared.core.events.split_events(
            actual_events,
            UserUttered,
            additional_splitting_conditions=additional_splitting_conditions,
            include_splitting_event=True,
        )

        if len(step_events) < 2:
            structlogger.error(
                "e2e_test_runner.run_assertions.user_message_not_found",
                message=f"User message '{step.text}' was not found in "
                f"the actual events. The user message "
                f"properties which were searched: "
                f"{additional_splitting_conditions}",
            )
            return [], []

        post_step_events = step_events[1]
        prior_events = step_events[0]

        # subset of events until the next user message
        turn_events = []
        for event in post_step_events:
            # we reached the next user message
            if isinstance(event, UserUttered) and step.text != event.text:
                break

            turn_events.append(event)

        return turn_events, prior_events

    @staticmethod
    def _slice_turn_events(
        step: TestStep,
        matching_event: Event,
        turn_events: List[Event],
        prior_events: List[Event],
    ) -> Tuple[List[Event], List[Event]]:
        """Slices the turn events when assertion order is enabled."""
        if not step.assertion_order_enabled:
            return turn_events, prior_events

        if not matching_event:
            return turn_events, prior_events

        matching_event_index = turn_events.index(matching_event)
        if matching_event_index + 1 < len(turn_events):
            prior_events += turn_events[: matching_event_index + 1]
            turn_events = turn_events[matching_event_index + 1 :]

        return turn_events, prior_events

    async def run_assertions(
        self,
        sender_id: str,
        test_case: TestCase,
        input_metadata: Optional[List[Metadata]],
    ) -> TestResult:
        """Runs the assertions defined in the test case."""
        tracker = await self.agent.processor.get_tracker(sender_id)  # type: ignore[union-attr]

        assertion_failure = None
        assertion_failure_found = False
        input_metadata = input_metadata if input_metadata else []

        for index, step in enumerate(test_case.steps):
            if not step.assertions:
                structlogger.debug(
                    "e2e_test_runner.run_assertions.no_assertions.skipping_step",
                    step=step,
                )
                continue

            additional_splitting_conditions = self._get_additional_splitting_conditions(
                step, input_metadata, tracker, test_case
            )

            turn_events, prior_events = self._get_current_user_turn_and_prior_events(
                tracker, additional_splitting_conditions, step
            )

            if not turn_events:
                return TestResult(
                    pass_status=False,
                    test_case=test_case,
                    difference=[],
                    error_line=step.line,
                    assertion_failure=None,
                )

            for assertion in step.assertions:
                structlogger.debug(
                    "e2e_test_runner.run_assertions.running_assertion",
                    test_case_name=test_case.name,
                    step_text=step.text,
                    assertion_type=assertion.type(),
                )

                assertion_order_error_msg = ""

                if step.assertion_order_enabled:
                    assertion_order_error_msg = (
                        " You have enabled assertion order, "
                        "you should check the order in which the "
                        "assertions are listed for this user step."
                    )

                assertion_failure, matching_event = assertion.run(
                    turn_events,
                    prior_events=prior_events,
                    assertion_order_error_message=assertion_order_error_msg,
                    llm_judge_config=self.llm_judge_config,
                    step_text=step.text,
                    step_index=index,
                )

                if assertion_failure:
                    assertion_failure_found = True
                    structlogger.debug(
                        "e2e_test_runner.run_assertions.assertion_failure_found",
                        test_case_name=test_case.name,
                        error_line=assertion_failure.error_line,
                    )
                    break

                turn_events, prior_events = self._slice_turn_events(
                    step, matching_event, turn_events, copy.deepcopy(prior_events)
                )

            if assertion_failure_found:
                # don't continue with the next steps if an assertion failed
                break

        return TestResult(
            pass_status=not assertion_failure,
            test_case=test_case,
            difference=[],
            error_line=assertion_failure.error_line if assertion_failure else None,
            assertion_failure=assertion_failure,
        )

    @classmethod
    def _resolve_successful_bot_utter(
        cls,
        expected_result: TestStep,
        test_response: ActualStepOutput,
    ) -> str:
        """Returns the diff text for a successful bot utter test step."""
        for event in test_response.bot_uttered_events:
            if expected_result.matches_event(event):
                # remove the event that is already matched so
                # that we dont compare against it again
                test_response.remove_bot_uttered_event(event)
                if expected_result.text is not None:
                    text = f"{expected_result.actor}: {expected_result.text}"
                elif expected_result.template is not None:
                    text = f"{expected_result.actor}: {expected_result.template}"
                break

        return text

    @classmethod
    def _resolve_successful_set_slot(
        cls,
        expected_result: TestStep,
        test_response: ActualStepOutput,
    ) -> str:
        """Returns the diff text for a successful set slot test step."""
        slot_name = expected_result.get_slot_name()
        text = f"slot_was_set: {slot_name}"

        for event in test_response.slot_set_events:
            if expected_result.matches_event(event):
                # remove the event that is already matched so
                # that we dont compare against it again
                test_response.remove_slot_set_event(event)
                if event.value is not None:
                    text += f": {event.value} ({type(event.value).__name__})"
                break

        return text

    @classmethod
    def _resolve_successful_slot_was_not_set(cls, expected_result: TestStep) -> str:
        """Returns the diff text for a successful slot was not set test step."""
        slot_name = expected_result.get_slot_name()
        slot_value = expected_result.get_slot_value()
        text = f"slot_was_not_set: {slot_name}"
        if expected_result.is_slot_instance_dict():
            text += f": {slot_value}"

        return text

    @classmethod
    def _find_first_non_matched_utterance(
        cls,
        test_response: ActualStepOutput,
        bot_utter_test_steps: List[TestStep],
    ) -> Optional[BotUttered]:
        """Finds the first non matched utterance in events (if any)."""
        bot_events = test_response.bot_uttered_events
        matching_events = []
        for event in bot_events:
            for test_step in bot_utter_test_steps:
                if test_step.matches_event(event):
                    matching_events.append(event)
                    break
        not_matching_events = [
            event for event in bot_events if event not in matching_events
        ]
        return not_matching_events[0] if not_matching_events else None

    @classmethod
    def _handle_fail_diff(
        cls,
        failed_step: TestStep,
        test_response: ActualStepOutput,
        bot_utter_test_steps: List[TestStep],
    ) -> Tuple[str, str]:
        """Handles generating the diff text for a failed test step."""
        diff_test_text = ""
        diff_actual_text = ""

        if failed_step.text is not None:
            diff_test_text = f"{failed_step.actor}: {failed_step.text}"
            diff_actual_text = NO_RESPONSE
            event: Optional[Union[BotUttered, SlotSet]] = (
                cls._find_first_non_matched_utterance(
                    test_response, bot_utter_test_steps
                )
            )
            if event and isinstance(event, BotUttered):
                test_response.remove_bot_uttered_event(event)
                diff_actual_text = f"bot: {event.text}"

        elif failed_step.template is not None:
            diff_test_text = f"{failed_step.actor}: {failed_step.template}"
            diff_actual_text = NO_RESPONSE
            event = cls._find_first_non_matched_utterance(
                test_response, bot_utter_test_steps
            )
            if event:
                test_response.remove_bot_uttered_event(event)
                diff_actual_text = f"bot: {event.metadata.get('utter_action')}"

        elif failed_step.slot_was_set:
            slot_name = failed_step.get_slot_name()

            diff_test_text = f"slot_was_set: {slot_name}"
            if failed_step.is_slot_instance_dict():
                slot_value = failed_step.get_slot_value()
                diff_test_text += f": {slot_value} ({type(slot_value).__name__})"

            diff_actual_text = NO_SLOT
            for event in test_response.slot_set_events:
                if slot_name == event.key:
                    diff_actual_text = (
                        f"slot_was_set: {event.key}: {event.value} "
                        f"({type(event.value).__name__})"
                    )
                    test_response.remove_slot_set_event(event)
                    break

        elif failed_step.slot_was_not_set:
            slot_name = failed_step.get_slot_name()

            diff_test_text = f"slot_was_not_set: {slot_name}"
            if failed_step.is_slot_instance_dict():
                slot_value = failed_step.get_slot_value()
                diff_test_text += f": {slot_value}"

            diff_actual_text = NO_SLOT
            for event in test_response.slot_set_events:
                if slot_name == event.key:
                    diff_actual_text = (
                        f"slot_was_set: {event.key}: {event.value} "
                        f"({type(event.value).__name__})"
                    )
                    test_response.remove_slot_set_event(event)
                    break

        return diff_test_text, diff_actual_text

    @classmethod
    def _select_bot_utter_turns(
        cls,
        test_turns: TEST_TURNS_TYPE,
        start_index: int,
    ) -> List[TestStep]:
        """Selects the TestSteps from the test turns that match BotUttered events."""
        bot_utter_turns = []
        for index in range(start_index, len(test_turns) - 1):
            test_turn = test_turns[index]
            if isinstance(test_turn, TestStep):
                if test_turn.text is not None or test_turn.template is not None:
                    bot_utter_turns.append(test_turn)
            elif isinstance(test_turn, ActualStepOutput):
                break

        return bot_utter_turns

    @classmethod
    def human_readable_diff(
        cls,
        test_turns: TEST_TURNS_TYPE,
        fail_positions: List[Tuple[TestFailure, int]],
    ) -> List[str]:
        """Returns a human readable diff of the test case and the actual conversation.

        Given an ordered list of test steps and actual conversation events, this
        method will return a human readable diff of the two.
        The diff uses difflib to compare the two lists and will highlight
        differences between the two in a pytest style diff.

        Args:
            test_turns: The transcript of test cases and events.
            fail_positions: The positions of the test failures.

        Returns:
            The human readable diff.
        """
        actual_transcript = []
        expected_transcript = []
        failure_points = {position for _, position in fail_positions}
        # This will only be used when the TestCase is not started
        # with a user step
        latest_response: ActualStepOutput = test_turns[-1]  # type: ignore[assignment]

        for index in range(max(failure_points) + 1):
            test_result = test_turns[index]
            if index in failure_points:
                diff_test_text, diff_actual_text = cls._handle_fail_diff(
                    test_result,
                    latest_response,
                    cls._select_bot_utter_turns(test_turns, index),
                )  # test_result can only be TestStep in failure_points
                actual_transcript.append(diff_actual_text)
                expected_transcript.append(diff_test_text)
                continue

            if isinstance(test_result, TestStep):
                if test_result.text is not None or test_result.template is not None:
                    diff_test_text = cls._resolve_successful_bot_utter(
                        expected_result=test_result,
                        test_response=latest_response,
                    )

                if test_result.slot_was_set:
                    diff_test_text = cls._resolve_successful_set_slot(
                        expected_result=test_result,
                        test_response=latest_response,
                    )

                if test_result.slot_was_not_set:
                    diff_test_text = cls._resolve_successful_slot_was_not_set(
                        expected_result=test_result,
                    )

            elif isinstance(test_result, ActualStepOutput):
                latest_response = test_result
                event = test_result.get_user_uttered_event()
                if event:
                    diff_test_text = f"{test_result.actor}: {event.text}"
                else:
                    raise RasaException(
                        f"Bot did not catch user "
                        f"event: {test_result.actor}: {test_result.text}."
                    )
            else:
                raise ValueError(f"Unexpected test result type: {type(test_result)}")

            # test passed in these cases so it's the same
            actual_transcript.append(diff_test_text)
            expected_transcript.append(diff_test_text)

        return list(difflib.ndiff(actual_transcript, expected_transcript))

    @classmethod
    def find_test_failures(
        cls,
        test_turns: TEST_TURNS_TYPE,
        test_case: Optional[TestCase],
        last_user_step_position: int = 0,
    ) -> List[Tuple[TestFailure, int]]:
        """Finds the test failures in the transcript.

        Args:
            test_turns (TEST_TURNS_TYPE): The transcript of test cases and events.
            test_case (Optional[TestCase]): The test case.
            last_user_step_position (int): The start position in the test turns.

        Returns:
            The test failures or an empty list if there is no test failure.
        """
        # This will only be used when the TestCase is not started
        # with a user step
        latest_response_event = test_turns[-1]
        if not isinstance(latest_response_event, ActualStepOutput):
            raise ValueError(
                f"Expected ActualStepOutput, got {type(latest_response_event)}"
            )
        latest_response: ActualStepOutput = latest_response_event
        failures = []
        match = None
        for position in range(last_user_step_position, len(test_turns) - 1):
            if position not in test_turns:
                continue
            turn_value = test_turns[position]
            if isinstance(turn_value, ActualStepOutput):
                latest_response = turn_value
            elif isinstance(turn_value, TestStep):
                if turn_value.text is not None or turn_value.template is not None:
                    match = cls._does_match_exist(
                        latest_response.bot_uttered_events, turn_value
                    )
                if turn_value.slot_was_set:
                    match = cls._does_match_exist(
                        latest_response.slot_set_events, turn_value
                    )
                if turn_value.slot_was_not_set:
                    match = not cls._does_match_exist(
                        latest_response.slot_set_events, turn_value
                    )
                if not match:
                    failures.append((TestFailure(test_case, turn_value.line), position))
            else:
                raise ValueError(f"Unexpected turn value type: {type(turn_value)}")

        return failures

    @classmethod
    def _does_match_exist(
        cls,
        actual_events: Optional[List[Union[UserUttered, BotUttered, SlotSet]]],
        expected: TestStep,
    ) -> bool:
        if not actual_events:
            return False

        match = False
        for event in actual_events:
            if expected.matches_event(event):
                return True
        return match

    async def set_up_fixtures(
        self,
        fixtures: List[Fixture],
        sender_id: Text,
    ) -> None:
        """Sets slots in the tracker as defined by the input fixtures.

        Args:
            fixtures: List of `Fixture` objects.
            sender_id: The conversation id.
        """
        if not fixtures:
            return
        if not self.agent.processor:
            return

        tracker = await self.agent.processor.fetch_tracker_with_initial_session(
            sender_id, output_channel=CollectingOutputChannel()
        )

        for fixture in fixtures:
            for slot_name, slot_value in fixture.slots_set.items():
                tracker.update(SlotSet(slot_name, slot_value))

        await self.agent.tracker_store.save(tracker)

    @staticmethod
    def filter_fixtures_for_test_case(
        test_case: TestCase, fixtures: List[Fixture]
    ) -> List[Fixture]:
        """Filters the input fixtures for the input test case.

        Args:
            test_case: The test case.
            fixtures: The fixtures.

        Returns:
        The filtered fixtures.
        """
        return list(
            filter(
                lambda fixture: test_case.fixture_names
                and fixture.name in test_case.fixture_names,
                fixtures,
            )
        )

    @staticmethod
    def filter_metadata_for_input(
        metadata_name: Optional[Text], test_suite_metadata: List[Metadata]
    ) -> Optional[Metadata]:
        """Filters the test suite metadata for a metadata name.

        Args:
            metadata_name: The test case or user step metadata name.
            test_suite_metadata: The top level list of all metadata definitions.

        Returns:
            The filtered metadata.
        """
        if not metadata_name:
            return None

        filtered_metadata = list(
            filter(
                lambda metadata: metadata_name and metadata.name == metadata_name,
                test_suite_metadata,
            )
        )

        if not filtered_metadata:
            structlogger.warning(
                "e2e_test_runner.filter_metadata_for_input",
                message=f"Metadata '{metadata_name}' is not defined in the input "
                f"metadata.",
            )
            return None

        return filtered_metadata[0]

    async def run_tests(
        self,
        input_test_cases: List[TestCase],
        input_fixtures: List[Fixture],
        fail_fast: bool = False,
        **kwargs: Any,
    ) -> List["TestResult"]:
        """Runs the test cases.

        Args:
            input_test_cases: Input test cases.
            input_fixtures: Input fixtures.
            fail_fast: Whether to fail fast.
            **kwargs: Additional arguments which are passed here.

        Returns:
            List of test results.
        """
        results = []
        input_metadata = kwargs.get("input_metadata", None)

        # telemetry call for tracking test runs
        track_e2e_test_run(input_test_cases, input_fixtures, input_metadata)

        # Validate and convert all fixtures upfront before running any tests
        # This ensures we fail fast on validation errors rather than running
        # some tests before hitting an invalid fixture, and stores converted
        # values in fixtures to avoid repeated validation/conversion
        input_fixtures = self._validate_and_convert_all_fixtures(input_fixtures)

        for test_case in input_test_cases:
            test_case_name = test_case.name.replace(" ", "_")
            # Add the name of the file and the current test case name being
            # executed in order to properly retrieve stub custom action
            if self.agent.endpoints and self.agent.endpoints.action:
                self.agent.endpoints.action.kwargs[TEST_FILE_NAME] = Path(
                    test_case.file
                ).name
                self.agent.endpoints.action.kwargs[TEST_CASE_NAME] = test_case_name

            sender_id = self.generate_sender_id(test_case.name)
            _, test_result = await self._process_test_case(
                test_case, sender_id, input_fixtures, input_metadata
            )

            results.append(test_result)

            coverage = kwargs.get("coverage", False)
            if coverage:
                tracker = await self.agent.tracker_store.retrieve(sender_id)
                if tracker:
                    # Get flows for slot-to-step mapping (to track collect steps)
                    flows = None
                    if self.agent.processor:
                        flows = await self.agent.processor.get_flows()
                    test_result.tested_paths, test_result.tested_commands = (
                        self._get_tested_flow_paths_and_commands(
                            tracker.events, test_result, flows
                        )
                    )

            if fail_fast and not test_result.pass_status:
                break

        return results

    async def _run_test_case(
        self,
        sender_id: str,
        input_fixtures: List[Fixture],
        input_metadata: Optional[List[Metadata]],
        test_case: TestCase,
    ) -> TEST_TURNS_TYPE:
        collector = CollectingOutputChannel()

        if input_fixtures:
            test_fixtures = self.filter_fixtures_for_test_case(
                test_case, input_fixtures
            )
            await self.set_up_fixtures(test_fixtures, sender_id)

        test_case_metadata = None
        if input_metadata:
            test_case_metadata = self.filter_metadata_for_input(
                test_case.metadata_name, input_metadata
            )

        return await self.run_prediction_loop(
            collector,
            test_case.steps,
            sender_id,
            test_case_metadata,
            input_metadata,
        )

    @staticmethod
    def generate_sender_id(test_case_name: str) -> str:
        # add timestamp suffix to ensure sender_id is unique
        return f"{test_case_name}_{datetime.now()}"

    def _validate_and_convert_all_fixtures(
        self, fixtures: List[Fixture]
    ) -> List[Fixture]:
        """Validate and convert all fixtures upfront before running any tests.

        This ensures we fail fast on validation errors (e.g., invalid mocked_datetime)
        rather than running some tests before hitting an invalid fixture.
        Collects all validation errors and reports them in a single ValidationError.
        Also converts mocked_datetime values to ISO format and stores them in fixtures
        to avoid repeated validation/conversion.

        Args:
            fixtures: List of all fixtures to validate and convert.

        Returns:
            List of fixtures with converted mocked_datetime values.

        Raises:
            ValidationError: If any fixture has invalid mocked_datetime values.
                The error message includes all invalid fixtures.
        """
        validation_errors: List[
            Tuple[str, str]
        ] = []  # List of (fixture_name, mocked_datetime_value)
        converted_fixtures = []

        # Single pass: validate and convert fixtures
        for fixture in fixtures:
            converted_slots = fixture.slots_set.copy()
            for slot_name, slot_value in fixture.slots_set.items():
                if slot_name == MOCKED_DATETIME_SLOT:
                    try:
                        # Validate and convert to ISO format
                        converted_value = self._get_validated_mocked_datetime(
                            slot_value
                        )
                        converted_slots[slot_name] = converted_value
                    except ValidationError:
                        # Collect the error for later reporting
                        validation_errors.append((fixture.name, str(slot_value)))

            # Create new fixture with converted values (even if validation failed,
            # we'll raise an error after processing all fixtures)
            converted_fixture = dataclasses.replace(fixture, slots_set=converted_slots)
            converted_fixtures.append(converted_fixture)

        # If there are any validation errors, raise a single comprehensive error
        if validation_errors:
            valid_formats = [
                "YYYY-MM-DDTHH:MM:SS±HH:MM  e.g. '2024-01-15T14:30:00+05:30'",
                "YYYY-MM-DDTHH:MM:SS±HHMM   e.g. '2024-01-15T14:30:00+0530'",
                "YYYY-MM-DD HH:MM:SS        e.g. '2024-01-15 14:30:00'",
                "YYYY-MM-DDTHH:MM:SS        e.g. '2024-01-15T14:30:00'",
                "YYYY-MM-DD                 e.g. '2024-01-15'",
            ]

            error_details = "\n".join(
                f"{i + 1}. Fixture - `{fixture_name}`, mocked_datetime value: "
                f"`{mocked_datetime_value}`"
                for i, (fixture_name, mocked_datetime_value) in enumerate(
                    validation_errors
                )
            )

            raise ValidationError(
                code="e2e_test_runner.validate_mocked_datetime.invalid_value_format",
                event_info=(
                    "Unable to convert to a valid datetime. Invalid `mocked_datetime` "
                    "value present in the following fixtures."
                    + f"\n\n{error_details}\n\n"
                    + "Accepted formats include:\n"
                    + "\n".join(f"  * {fmt}" for fmt in valid_formats)
                ),
            )

        return converted_fixtures

    def _get_validated_mocked_datetime(
        self, mocked_datetime_value: Any
    ) -> Optional[str]:
        """Validates and converts mocked_datetime to ISO 8601 format.

        Args:
            mocked_datetime_value: The value of the mocked_datetime slot.
                Expected to be a string from YAML fixtures, or None.

        Returns:
            An ISO 8601 format string (timezone-aware) or None.
            Example: '2024-01-15T14:30:45+00:00'

        Raises:
            ValidationError: If the mocked_datetime value cannot be converted
                to a datetime object.
        """
        if mocked_datetime_value is None:
            return None

        if not isinstance(mocked_datetime_value, str):
            # YAML fixtures only provide strings, so other types are invalid
            raise ValidationError(
                code="e2e_test_runner.validate_mocked_datetime.invalid_value_type",
                event_info="Unable to convert to a valid datetime.",
            )

        valid_datetime_formats = {
            "%Y-%m-%dT%H:%M:%S%z": True,  # already timezone-aware
            "%Y-%m-%d %H:%M:%S": False,  # needs UTC
            "%Y-%m-%dT%H:%M:%S": False,  # needs UTC
            "%Y-%m-%d": False,  # needs UTC
        }

        for datetime_format, timezone_aware in valid_datetime_formats.items():
            try:
                parsed = datetime.strptime(mocked_datetime_value, datetime_format)
                # Ensure timezone-aware
                if not timezone_aware:
                    parsed = parsed.replace(tzinfo=timezone.utc)
                # Return ISO format string
                return parsed.isoformat()
            except ValueError:
                continue

        # If we get here, the conversion failed. i.e. all parsing attempts fail
        raise ValidationError(
            code="e2e_test_runner.validate_mocked_datetime.invalid_value_format",
            event_info="Unable to convert to a valid datetime.",
        )

    async def _process_test_case(
        self,
        test_case: TestCase,
        sender_id: str,
        input_fixtures: List[Fixture],
        input_metadata: Optional[List[Metadata]],
    ) -> Tuple[TEST_TURNS_TYPE, TestResult]:
        """Runs a single test case and returns the test turns and result."""
        test_turns = await self._run_test_case(
            sender_id, input_fixtures, input_metadata, test_case
        )

        if not test_case.uses_assertions():
            test_result = self.generate_test_result(test_turns, test_case)
        else:
            test_result = await self.run_assertions(
                sender_id, test_case, input_metadata
            )

        return test_turns, test_result

    async def run_tests_for_fine_tuning(
        self,
        input_test_cases: List[TestCase],
        input_fixtures: List[Fixture],
        input_metadata: Optional[List[Metadata]],
    ) -> List[Conversation]:
        """Runs the test cases for fine-tuning.

        Converts passing test cases into conversation objects containing the
        prompts and llm commands per user message.

        Args:
            input_test_cases: Input test cases.
            input_fixtures: Input fixtures.
            input_metadata: Input metadata.

        Returns:
            List of conversations.
        """
        import rasa.llm_fine_tuning.annotation_module

        conversations = []

        for i in tqdm(range(len(input_test_cases))):
            test_case = input_test_cases[i]
            sender_id = self.generate_sender_id(test_case.name)

            test_turns, test_result = await self._process_test_case(
                test_case, sender_id, input_fixtures, input_metadata
            )

            if not test_result.pass_status:
                structlogger.warning(
                    "annotation_module.skip_test_case.failing_e2e_test",
                    test_case=test_case.name,
                    file=test_case.file,
                )
                continue

            tracker = await self.agent.tracker_store.retrieve(sender_id)
            conversation = rasa.llm_fine_tuning.annotation_module.generate_conversation(
                test_turns, test_case, tracker, test_case.uses_assertions()
            )

            if conversation:
                conversations.append(conversation)

        return conversations

    async def run_tests_to_convert_tests_to_du_tests(
        self,
        input_test_cases: List[TestCase],
        input_fixtures: List[Fixture],
        input_metadata: Optional[List[Metadata]],
        converting_method: Callable[
            [TEST_TURNS_TYPE, TestCase, bool, bool],
            Optional[DialogueUnderstandingTestCase],
        ],
    ) -> Tuple[
        List[DialogueUnderstandingTestCase], List[DialogueUnderstandingTestCase]
    ]:
        """Runs the test cases to convert them into dialogue understanding tests.

        Converts test cases into dialogue understanding test cases.

        Args:
            input_test_cases: Input test cases.
            input_fixtures: Input fixtures.
            input_metadata: Input metadata.
            converting_method: The method to convert the e2e test case into a
              dialogue understanding test case.

        Returns:
            List of ready dialogue understanding test cases and list of
            dialogue understanding test cases to review.
        """
        ready_du_test_cases = []
        to_review_du_test_cases = []

        for i in tqdm(range(len(input_test_cases))):
            test_case = input_test_cases[i]
            sender_id = self.generate_sender_id(test_case.name)

            test_turns, test_result = await self._process_test_case(
                test_case, sender_id, input_fixtures, input_metadata
            )

            du_test_case = converting_method(
                test_turns,
                test_case,
                test_case.uses_assertions(),
                test_result.pass_status,
            )

            if du_test_case:
                if test_result.pass_status:
                    ready_du_test_cases.append(du_test_case)
                else:
                    to_review_du_test_cases.append(du_test_case)

        return ready_du_test_cases, to_review_du_test_cases

    @staticmethod
    def _action_server_is_reachable(
        endpoints: AvailableEndpoints, module: str = "e2e_test_runner"
    ) -> None:
        """Calls the action server health endpoint."""
        if not endpoints.action:
            structlogger.debug(
                f"{module}._action_server_is_reachable",
                message="No action endpoint configured. Skipping the health check "
                "of the action server.",
            )
            return

        if endpoints.action.actions_module:
            structlogger.debug(
                f"{module}._action_server_is_reachable",
                message="Rasa server is configured to run custom actions directly. "
                "Skipping the health check of the action server.",
            )
            return

        if not endpoints.action.url:
            structlogger.debug(
                f"{module}._action_server_is_reachable",
                message="Action endpoint URL is not defined in the endpoint "
                "configuration.",
            )
            return

        structlogger.debug(
            f"{module}._action_server_is_reachable",
            message="Detected action URL in the endpoint configuration.\n"
            f"Action Server URL: {endpoints.action.url}\n"
            "Sending a health request to the action endpoint.",
        )
        url = urlparse(endpoints.action.url)
        # replace /<path> with just /health
        url = url._replace(path="/health").geturl()  # type: ignore[assignment]
        try:
            response = requests.get(url, timeout=3)
        except requests.exceptions.ConnectionError as error:
            raise RasaException(
                "Action endpoint could not be reached. "
                "Actions server URL is defined in your endpoint configuration as "
                f"'{endpoints.action.url}'.\n"
                "Please make sure your action server is running and properly "
                "configured. Since running tests without a action server may "
                f"lead to unpredictable results.\n{error}"
            )

        if response.status_code != 200:
            raise RasaException(
                "Action endpoint is responding, but health status responded with "
                f"code {response.status_code}. Make sure your action server"
                " is properly configured and that the '/health' endpoint is available."
            )

        structlogger.debug(
            f"{module}._action_server_is_reachable",
            message="Action endpoint has responded successfully.\n"
            f"Response message: {response.text}\n"
            f"Response status code: {response.status_code}.",
        )

    def _get_tested_flow_paths_and_commands(
        self,
        events: List[Event],
        test_result: TestResult,
        flows: Optional["FlowsList"] = None,
    ) -> Tuple[Optional[List[FlowPath]], Dict[str, Dict[str, int]]]:
        """Extract tested paths and commands from dialog events.

        A flow path consists of bot utterances, custom actions, and inferred
        step executions (call/link steps, collect steps with prefilled slots).

        Args:
            events: The list of dialog events.
            test_result: The result of the test incl. the pass status.
            flows: Optional flow definitions for slot-to-step mapping.

        Returns:
            Tuple[flow_paths: Optional[List[FlowPath]], tested_commands:
            Dict[str, Dict[str, int]]], where tested_commands is a
            dictionary like
            {"flow1": {"set slot": 5, "clarify": 1}, "flow2": {"set slot": 3}}
        """
        tested_paths = []
        # we want to create a flow path per flow the e2e test covers
        # as an e2e test can cover multiple flows, we might end up creating
        # multiple flow paths
        _tested_commands: DefaultDict[str, DefaultDict[str, int]] = defaultdict(
            lambda: defaultdict(int)
        )
        flow_paths_stack: list[FlowPath] = []

        # Build slot-to-step mapping for tracking collect steps with prefilled slots
        slot_to_collect_step = self._build_slot_to_collect_step_mapping(flows)

        for event in events:
            if isinstance(event, FlowStarted) and not event.flow_id.startswith(
                RASA_DEFAULT_FLOW_PATTERN_PREFIX
            ):
                # When a child flow starts due to a call/link step,
                # mark the parent's call/link step as visited using the metadata
                # that was added by the flow executor.
                if (
                    flow_paths_stack
                    and PARENT_FLOW_ID_METADATA_KEY in event.metadata
                    and PARENT_STEP_ID_METADATA_KEY in event.metadata
                ):
                    parent_flow_id = event.metadata[PARENT_FLOW_ID_METADATA_KEY]
                    parent_step_id = event.metadata[PARENT_STEP_ID_METADATA_KEY]
                    # Find the parent flow path on the stack and add the call/link step
                    for flow_path in reversed(flow_paths_stack):
                        if flow_path.flow == parent_flow_id:
                            call_step_node = PathNode(
                                step_id=parent_step_id,
                                flow=parent_flow_id,
                            )
                            # Only add if not already present
                            if call_step_node not in flow_path.nodes:
                                flow_path.nodes.append(call_step_node)
                            break

                flow_paths_stack.append(FlowPath(event.flow_id))

            elif (
                isinstance(event, FlowCompleted)
                and len(flow_paths_stack) > 0
                and event.flow_id == flow_paths_stack[-1].flow
            ):
                # flow path is completed as the flow ended
                tested_paths.append(flow_paths_stack.pop())

            elif isinstance(event, BotUttered):
                if (
                    flow_paths_stack
                    and STEP_ID_METADATA_KEY in event.metadata
                    and ACTIVE_FLOW_METADATA_KEY in event.metadata
                ):
                    flow_paths_stack[-1].nodes.append(self._create_path_node(event))

            elif isinstance(event, ActionExecuted):
                # we are only interested in custom actions
                if (
                    flow_paths_stack
                    and self.agent.domain
                    and self.agent.domain.is_custom_action(event.action_name)
                    and STEP_ID_METADATA_KEY in event.metadata
                    and ACTIVE_FLOW_METADATA_KEY in event.metadata
                ):
                    flow_paths_stack[-1].nodes.append(self._create_path_node(event))

            elif isinstance(event, SlotSet):
                # Track collect steps with prefilled slots.
                # When a slot is set (e.g., from user message extraction), find the
                # collect step in the current flow that collects this slot.
                if flow_paths_stack and slot_to_collect_step:
                    current_flow_id = flow_paths_stack[-1].flow
                    slot_name = event.key
                    # Look up the collect step for this slot in the current flow
                    if (current_flow_id, slot_name) in slot_to_collect_step:
                        step_id = slot_to_collect_step[(current_flow_id, slot_name)]
                        collect_step_node = PathNode(
                            step_id=step_id,
                            flow=current_flow_id,
                        )
                        # Only add if not already present (avoid duplicates)
                        if collect_step_node not in flow_paths_stack[-1].nodes:
                            flow_paths_stack[-1].nodes.append(collect_step_node)

            # Time to gather tested commands
            elif isinstance(event, UserUttered):
                if event.parse_data and COMMANDS in event.parse_data:
                    commands = [
                        command["command"] for command in event.parse_data[COMMANDS]
                    ]
                    current_flow = (
                        flow_paths_stack[-1].flow if flow_paths_stack else "no_flow"
                    )
                    for command in commands:
                        _tested_commands[current_flow][command] += 1

        # It might be that an e2e test stops before a flow was completed.
        # Add the remaining flow paths to the tested paths list.
        while len(flow_paths_stack) > 0:
            tested_paths.append(flow_paths_stack.pop())

        # Convert _tested_commands to normal dicts
        tested_commands = {key: dict(value) for key, value in _tested_commands.items()}  # type: Dict[str, Dict[str, int]]

        return tested_paths, tested_commands

    @staticmethod
    def _create_path_node(event: Event) -> PathNode:
        flow_id = event.metadata[ACTIVE_FLOW_METADATA_KEY]
        step_id = event.metadata[STEP_ID_METADATA_KEY]
        return PathNode(step_id=step_id, flow=flow_id)

    @staticmethod
    def _build_slot_to_collect_step_mapping(
        flows: Optional["FlowsList"],
    ) -> Dict[Tuple[str, str], str]:
        """Build a mapping from (flow_id, slot_name) to collect step_id.

        This allows looking up which collect step corresponds to a given slot
        when the slot is set, so we can track collect steps even when the slot
        was prefilled (and thus no BotUttered event was emitted).

        Args:
            flows: The flow definitions.

        Returns:
            Dict mapping (flow_id, slot_name) to the collect step's step_id.
        """
        mapping: Dict[Tuple[str, str], str] = {}
        if not flows:
            return mapping

        for flow in flows.underlying_flows:
            for step in flow.get_collect_steps():
                # step.collect is the slot name being collected
                # step.id is the step's ID
                mapping[(flow.id, step.collect)] = step.id

        return mapping



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from rasa.builder.copilot.constants import ROLE_ASSISTANT, ROLE_USER
from rasa.shared.core.constants import DEFAULT_SLOT_NAMES
from rasa.shared.core.events import (
    ActionExecuted,
    BotUttered,
    FlowCompleted,
    FlowStarted,
    SlotSet,
    UserUttered,
)
from rasa.shared.core.trackers import DialogueStateTracker


class AssistantMessage(BaseModel):
    text: str


class UserMessage(BaseModel):
    text: str
    predicted_commands: List[str] = Field(default_factory=list)


class TrackerEvent(BaseModel):
    event: str
    data: Dict[str, Any] = Field(default_factory=dict)


class AssistantConversationTurn(BaseModel):
    user_message: Optional[UserMessage] = None
    assistant_messages: List[AssistantMessage] = Field(default_factory=list)
    context_events: List[TrackerEvent] = Field(default_factory=list)


class CurrentState(BaseModel):
    latest_message: Optional[str] = None
    active_flow: Optional[str] = None
    flow_stack: Optional[List[Dict[str, Any]]] = None
    slots: Optional[Dict[str, Any]] = None
    latest_action: Optional[str] = None
    followup_action: Optional[str] = None


class TrackerContext(BaseModel):
    conversation_turns: List[AssistantConversationTurn]
    current_state: CurrentState

    def to_openai_format(self) -> List[Dict[str, Any]]:
        """Convert the TrackerContext to a format suitable for OpenAI.

        Returns:
            A list of user and assistant messages in the format suitable for OpenAI.
        """
        messages: List[Dict[str, Any]] = []
        for turn in self.conversation_turns:
            if turn.user_message:
                messages.append({"role": ROLE_USER, "content": turn.user_message.text})
            for message in turn.assistant_messages:
                messages.append({"role": ROLE_ASSISTANT, "content": message.text})
        return messages

    @property
    def formatted_conversation_turns(self) -> List[Dict[str, Any]]:
        """Conversation turns in a structured format (USER / ASSISTANT / other events).

        Same data as conversation_turns but as a dict with key
        "conversation_history" and turns that clearly separate user message,
        assistant messages, and other tracker events.

        Example:
            [
                {
                    "turn_id": 1,
                    "USER": {
                        "text": "I want to transfer money",
                        "predicted_commands": ["start flow", "set slot", ...]
                    },
                    "ASSISTANT": [
                        {"text": "How much would you like to transfer?"}
                    ],
                    "other_tracker_events": [
                        {
                            "event": "action_executed",
                            "data": {"action_name": "action_ask_amount"}
                        },
                        {
                            "event": "slot_set",
                            "data": {
                                "slot_name": "amount_of_money",
                                "slot_value": 100,
                            },
                        }
                    ]
                }
            ]


        Returns:
            Dict with "conversation_history" key and list of turn dicts.
        """
        turn_id_key = "turn_id"
        user_turn_key = "USER"
        assistant_turn_key = "ASSISTANT"
        other_events_key = "other_tracker_events"
        formatted_turns: List[Dict[str, Any]] = []

        for turn_idx, turn in enumerate(self.conversation_turns, 1):
            turn_data: Dict[str, Any] = {turn_id_key: turn_idx}

            if turn.user_message:
                turn_data[user_turn_key] = turn.user_message.model_dump()

            if turn.assistant_messages:
                turn_data[assistant_turn_key] = [
                    assistant_message.model_dump()
                    for assistant_message in turn.assistant_messages
                ]

            if turn.context_events:
                turn_data[other_events_key] = [
                    event.model_dump() for event in turn.context_events
                ]

            formatted_turns.append(turn_data)

        return formatted_turns

    @classmethod
    def from_tracker(
        cls, tracker: Optional[DialogueStateTracker], max_turns: int = 10
    ) -> Optional["TrackerContext"]:
        """Convert a tracker to a TrackerContext."""
        if not tracker or not tracker.events:
            return None

        conversation_turns = cls._build_conversation_turns(tracker)
        conversation_turns = conversation_turns[-max_turns:]
        current_state = cls._build_current_state(tracker)

        return cls(
            conversation_turns=conversation_turns,
            current_state=current_state,
        )

    @classmethod
    def _build_conversation_turns(
        cls, tracker: DialogueStateTracker
    ) -> List[AssistantConversationTurn]:
        """Build conversation turns from tracker events."""
        conversation_turns: List[AssistantConversationTurn] = []
        current_user_message: Optional[UserMessage] = None
        current_assistant_messages: List[AssistantMessage] = []
        current_context_events: List[TrackerEvent] = []

        for event in tracker.applied_events():
            if isinstance(event, UserUttered):
                # Save previous turn if exists and has content. A new turn starts with a
                # user message. However, since it's possible that "turn" started with an
                # assistant message, we save that turn without the user message.
                if (
                    current_user_message
                    or current_assistant_messages
                    or current_context_events
                ):
                    conversation_turns.append(
                        AssistantConversationTurn(
                            user_message=current_user_message,
                            assistant_messages=current_assistant_messages,
                            context_events=current_context_events,
                        )
                    )
                current_assistant_messages = []
                current_context_events = []

                # Start new turn

                # Fetch the predicted commands for the user message
                predicted_commands = (
                    [command.get("command") for command in event.commands]
                    if event.commands
                    else []
                )

                current_user_message = UserMessage(
                    text=event.text or "",
                    predicted_commands=predicted_commands,
                )

            # Assistant conversation turn can have multiple messages from the assistant
            elif isinstance(event, BotUttered):
                current_assistant_messages.append(
                    AssistantMessage(text=event.text or "")
                )

            # Handle non-user and non-assistant events. These are useful for more
            # adding context to the conversation turn.
            else:
                context_event = cls._process_tracker_event(event)
                if context_event:
                    current_context_events.append(context_event)

        # Add the final turn if there is one
        if current_user_message or current_assistant_messages:
            conversation_turns.append(
                AssistantConversationTurn(
                    user_message=current_user_message,
                    assistant_messages=current_assistant_messages,
                    context_events=current_context_events,
                )
            )

        return conversation_turns

    @staticmethod
    def _process_tracker_event(event: Any) -> Optional[TrackerEvent]:
        if isinstance(event, ActionExecuted):
            return TrackerEvent(
                event=ActionExecuted.type_name,
                data={"action_name": event.action_name, "confidence": event.confidence},
            )

        elif isinstance(event, SlotSet) and event.key not in DEFAULT_SLOT_NAMES:
            return TrackerEvent(
                event=SlotSet.type_name,
                data={"slot_name": event.key, "slot_value": event.value},
            )

        elif isinstance(event, FlowStarted):
            return TrackerEvent(
                event=FlowStarted.type_name,
                data={"flow_id": event.flow_id},
            )

        elif isinstance(event, FlowCompleted):
            return TrackerEvent(
                event=FlowCompleted.type_name,
                data={"flow_id": event.flow_id},
            )
        else:
            return None

    @classmethod
    def _build_current_state(cls, tracker: DialogueStateTracker) -> CurrentState:
        """Build the current state from the tracker."""
        latest_message = tracker.latest_message.text if tracker.latest_message else None
        latest_action = (
            tracker.latest_action.get("action_name") if tracker.latest_action else None
        )
        followup_action = tracker.followup_action
        flow_stack = tracker.stack.as_dict() if tracker.stack else None
        slots = cls._extract_non_default_slots(tracker)
        active_flow = tracker.active_flow

        return CurrentState(
            latest_message=latest_message,
            active_flow=active_flow,
            flow_stack=flow_stack,
            latest_action=latest_action,
            followup_action=followup_action,
            slots=slots,
        )

    @classmethod
    def _extract_non_default_slots(
        cls, tracker: DialogueStateTracker
    ) -> Optional[Dict[str, Any]]:
        """Extract non-default slot values from the tracker."""
        if not tracker.slots:
            return None

        return {
            k: str(v.value)
            for k, v in tracker.slots.items()
            if v is not None and k not in DEFAULT_SLOT_NAMES
        }



"""Models for completeness judge."""

import uuid
from typing import List

from pydantic import BaseModel, Field

from rasa.builder.copilot.models import ChatMessage
from rasa.builder.evaluator.content_processors.models import Claims
from rasa.builder.evaluator.dataset.models import DatasetEntry


class CompletenessJudgeInput(BaseModel):
    """Model representing the input for the completeness judge."""

    user_message: str = Field(..., description="The most recent user message")
    chat_history: List[ChatMessage] = Field(
        default_factory=list,
        description="Full copilot chat history for context",
    )
    claims: Claims = Field(..., description="Claims extracted from Copilot response")

    @classmethod
    def from_dataset_entry_and_claims(
        cls,
        dataset_entry: DatasetEntry,
        claims: Claims,
    ) -> "CompletenessJudgeInput":
        user_message = dataset_entry.input.message or ""
        chat_history = (
            dataset_entry.metadata.copilot_additional_context.copilot_chat_history
        )
        return cls(
            user_message=user_message,
            chat_history=chat_history,
            claims=claims,
        )


class UserRequestCompletenessVerdict(BaseModel):
    """Model for the completeness verdict for a part of the user request.

    Represents one distinct aspect or question within the user's request.
    """

    part_text: str = Field(
        ..., description="Text describing this atomic part of the user request"
    )
    addressing_claims_ids: List[uuid.UUID] = Field(
        default_factory=list,
        description="List of claim UUIDs that address this part of the request",
    )
    rationale: str = Field(
        ...,
        description=(
            "Explanation of how the claims address (or don't address) this part"
        ),
    )

    @property
    def is_covered(self) -> bool:
        """Check if this part of the user request is covered by any claims."""
        return len(self.addressing_claims_ids) > 0


class CompletenessJudgeResult(BaseModel):
    """Model for the completeness judge result."""

    verdicts: List[UserRequestCompletenessVerdict] = Field(
        ..., description="List of verdicts for each part of the user request"
    )
    overall_rationale: str = Field(
        ..., description="Overall rationale for the completeness assessment"
    )
    confidence: float = Field(
        ...,
        ge=0.0,
        le=1.0,
        description="Confidence in the completeness assessment (0-1)",
    )

    @property
    def covered_parts(self) -> List[UserRequestCompletenessVerdict]:
        """Get the covered parts of the user request."""
        return [verdict for verdict in self.verdicts if verdict.is_covered]

    @property
    def uncovered_parts(self) -> List[UserRequestCompletenessVerdict]:
        """Get the uncovered parts of the user request."""
        return [verdict for verdict in self.verdicts if not verdict.is_covered]



from functools import wraps
from typing import TYPE_CHECKING, Any, Callable

from rasa.builder import config
from rasa.builder.copilot.models import UsageStatistics
from rasa.builder.telemetry.langfuse.langfuse_compat import (
    is_langfuse_available,
    langfuse,
)
from rasa.builder.telemetry.langfuse.shared import (
    update_generation_span_with_usage_statistics,
)

if TYPE_CHECKING:
    from rasa.builder.document_retrieval.inkeep_document_retrieval import (
        InKeepDocumentRetrieval,
    )


class DocumentRetrievalLangfuseTelemetry:
    """Telemetry for document retrieval generation with Langfuse."""

    @staticmethod
    def trace_document_retrieval_generation(
        func: Callable[..., Any],
    ) -> Callable[..., Any]:
        """Custom decorator for tracing document retrieval generation with Langfuse.

        This decorator handles Langfuse tracing for document retrieval API calls
        by manually managing the generation span and updating it with usage statistics.
        """
        if not is_langfuse_available():
            return func

        @wraps(func)
        async def wrapper(
            self: "InKeepDocumentRetrieval",
            query: str,
            temperature: float,
            timeout: float,
        ) -> Any:
            langfuse_client = langfuse.get_client()

            with langfuse_client.start_as_current_generation(
                name=f"{self.__class__.__name__}.{func.__name__}",
                input={
                    "query": query,
                    "temperature": temperature,
                    "timeout": timeout,
                },
            ) as generation:
                # Call the original function
                response = await func(self, query, temperature, timeout)

                # Update the span with response content
                generation.update(
                    output=response,
                    model_parameters={
                        "temperature": str(temperature),
                        "timeout": str(timeout),
                    },
                )

                # Update usage statistics if available
                usage_statistics = UsageStatistics.from_chat_completion_response(
                    response,
                    input_token_price=config.COPILOT_INPUT_TOKEN_PRICE,
                    output_token_price=config.COPILOT_OUTPUT_TOKEN_PRICE,
                    cached_token_price=config.COPILOT_CACHED_TOKEN_PRICE,
                )
                if usage_statistics:
                    update_generation_span_with_usage_statistics(
                        generation, usage_statistics
                    )

                return response

        return wrapper



import asyncio
import json
import os
from contextlib import asynccontextmanager
from functools import lru_cache
from typing import Any, AsyncGenerator, List, Optional

import openai
import structlog
from openai.types.chat import ChatCompletion

from rasa.builder import config
from rasa.builder.copilot.constants import ROLE_USER
from rasa.builder.document_retrieval.constants import (
    INKEEP_API_KEY_ENV_VAR,
    INKEEP_DOCUMENT_RETRIEVAL_MODEL,
    INKEEP_RAG_RESPONSE_SCHEMA_PATH,
)
from rasa.builder.document_retrieval.models import Document
from rasa.builder.exceptions import DocumentRetrievalError
from rasa.builder.telemetry.langfuse.document_retrieval_langfuse_telemetry import (
    DocumentRetrievalLangfuseTelemetry,
)
from rasa.shared.utils.io import read_json_file

structlogger = structlog.get_logger()


@lru_cache(maxsize=1)
def _read_rag_schema() -> Any:
    return read_json_file(INKEEP_RAG_RESPONSE_SCHEMA_PATH)


class InKeepDocumentRetrieval:
    """Handles the document retrieval from InKeep AI."""

    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        self._rag_schema = _read_rag_schema()
        self._api_key = api_key or os.getenv(INKEEP_API_KEY_ENV_VAR)

    @property
    def api_key(self) -> str:
        """Resolve the correct API key/token based on proxy usage."""
        using_proxy = bool(config.HELLO_LLM_PROXY_BASE_URL)
        if using_proxy:
            if not config.RASA_PRO_LICENSE:
                raise DocumentRetrievalError(
                    "HELLO_LLM_PROXY_BASE_URL is set but RASA_PRO_LICENSE is missing. "
                    "Proxy requires a Rasa Pro license token for authentication."
                )
            return config.RASA_PRO_LICENSE

        if not self._api_key:
            raise DocumentRetrievalError(
                "INKEEP_API_KEY is missing. Provide it via env INKEEP_API_KEY "
                "or pass api_key= to InKeepDocumentRetrieval."
            )
        return self._api_key

    async def retrieve_documents(
        self, query: str, temperature: float = 0.0, timeout: float = 30.0
    ) -> List[Document]:
        """Retrieve relevant documents using InKeep AI based on the given query.

        Args:
            query: The search query
            temperature: Controls randomness in generation (0.0 for deterministic)
            timeout: Timeout for the API call

        Returns:
            List of Document objects containing retrieved content

        Raises:
            DocumentRetrievalError: When document retrieval fails due to:
                - Empty response from InKeep AI API
                - OpenAI API errors (authentication, rate limiting, etc.)
                - Request timeout
                - Unexpected errors during API communication
        """
        try:
            response = await self._call_inkeep_rag_api(
                query=query,
                temperature=temperature,
                timeout=timeout,
            )
            documents = self._parse_documents_from_response(response)
            return documents
        except DocumentRetrievalError as e:
            structlogger.error(
                "inkeep_document_retrieval.retrieve_documents.error",
                event_info="InKeep Document Retrieval: Error",
                query=query,
                error=str(e),
            )
            raise e

    @DocumentRetrievalLangfuseTelemetry.trace_document_retrieval_generation
    async def _call_inkeep_rag_api(
        self, query: str, temperature: float, timeout: float
    ) -> ChatCompletion:
        """Call InKeep AI RAG's API endpoint and return the response content.

        Args:
            query: The search query to send to InKeep
            temperature: Controls randomness in generation (0.0 for deterministic)
            timeout: Timeout for the API call

        Returns:
            The response content from InKeep AI. The response is made of the retrieved
            documents.

        Raises:
            LLMGenerationError: If the API call fails or returns invalid response
        """
        try:
            async with self._get_client() as client:
                response = await client.chat.completions.create(
                    model=INKEEP_DOCUMENT_RETRIEVAL_MODEL,
                    messages=[{"role": ROLE_USER, "content": query}],
                    temperature=temperature,
                    timeout=timeout,
                    response_format={
                        "type": "json_schema",
                        "json_schema": self._rag_schema,
                    },
                )

                if not response.choices[0].message.content:
                    structlogger.warning(
                        "inkeep_document_retrieval.empty_response",
                        event_info="InKeep AI returned an empty response. ",
                        query=query,
                        response_content=response.choices[0].message.content,
                    )
                    raise DocumentRetrievalError(
                        "InKeep Document Retrieval: Empty response"
                    )

                return response

        except openai.OpenAIError as e:
            structlogger.error(
                "inkeep_document_retrieval.call_inkeep_rag_api.api_error",
                event_info="InKeep Document Retrieval: API error",
                query=query,
                error=e,
            )
            raise DocumentRetrievalError(f"InKeep Document Retrieval: API error: {e}")
        except asyncio.TimeoutError as e:
            structlogger.error(
                "inkeep_document_retrieval.call_inkeep_rag_api.timeout_error",
                event_info="InKeep Document Retrieval: Timeout error",
                query=query,
                error=e,
            )
            raise DocumentRetrievalError(f"InKeep AI request timed out: {e}")
        except Exception as e:
            structlogger.error(
                "inkeep_document_retrieval.call_inkeep_rag_api.error",
                event_info="InKeep Document Retrieval: Error",
                query=query,
                error=e,
            )
            raise DocumentRetrievalError(
                f"InKeep Document Retrieval: Unexpected error: {e}"
            )

    @asynccontextmanager
    async def _get_client(self) -> AsyncGenerator[openai.AsyncOpenAI, None]:
        """Get or create client that handles the API calls to InKeep AI."""
        # Ensure trailing slash to match client expectations/tests
        base_url = f"{config.INKEEP_BASE_URL.rstrip('/')}/"
        client = openai.AsyncOpenAI(
            base_url=base_url,
            api_key=self.api_key,
            default_headers={
                config.DEPLOYMENT_STACK_HEADER_NAME: config.DEPLOYMENT_STACK
            },
        )
        structlogger.debug("inkeep_document_retrieval._get_client", base_url=base_url)

        try:
            yield client
        except Exception as e:
            structlogger.error(
                "inkeep_document_retrieval.client_error",
                event_info="InKeep Document Retrieval: Client error",
                error=str(e),
            )
            raise

    def _parse_documents_from_response(
        self, response: ChatCompletion
    ) -> List[Document]:
        """Parse the InKeep AI response into Document objects.

        Args:
            response: ChatCompletion response from InKeep AI's RAG model.

        Returns:
            List of Document objects. Empty list is returned if the response is empty
            or if the response is invalid.
        """
        try:
            content = response.choices[0].message.content
            if not content:
                return []

            response_data = json.loads(content)
            documents = []

            for item in response_data.get("content", []):
                try:
                    document = Document.from_inkeep_rag_response(item)
                    documents.append(document)
                except Exception as e:
                    structlogger.warning(
                        "inkeep_document_retrieval.invalid_document_skipped",
                        event_info=(
                            "InKeep Document Retrieval: Invalid document structure "
                            "skipped. Returning empty list for this item."
                        ),
                        error=str(e),
                        item=item,
                    )
                    # Continue processing other items, skip this invalid one
                    continue

            return documents

        except json.JSONDecodeError as e:
            structlogger.warning(
                "inkeep_document_retrieval.parse_documents_from_response"
                ".parse_response_failed",
                event_info=(
                    "InKeep Document Retrieval: Parse response failed. "
                    "Returning empty list.",
                ),
                error=str(e),
            )
            return []
        except Exception as e:
            structlogger.error(
                "inkeep_document_retrieval.parse_documents_from_response.error",
                event_info=(
                    "InKeep Document Retrieval: Parse response error. "
                    "Returning empty list.",
                ),
                error=str(e),
            )
            return []



import argparse
import asyncio
import sys
from pathlib import Path
from typing import Any, Dict, List, Type, cast

import structlog

import rasa.cli.utils
import rasa.shared.utils.cli
import rasa.shared.utils.io
import rasa.shared.utils.yaml
from rasa.cli import SubParsersAction
from rasa.cli.arguments.default_arguments import (
    add_endpoint_param,
    add_model_param,
    add_remote_storage_param,
    add_sub_agents_param,
)
from rasa.cli.e2e_test import (
    RASA_PRO_BETA_FINE_TUNING_RECIPE_ENV_VAR_NAME,
    read_test_cases,
    validate_model_path,
)
from rasa.core.config.available_endpoints import AvailableEndpoints
from rasa.core.config.configuration import Configuration
from rasa.core.exceptions import AgentNotReady
from rasa.dialogue_understanding.generator.llm_based_command_generator import (
    LLMBasedCommandGenerator,
)
from rasa.dialogue_understanding.generator.multi_step.multi_step_llm_command_generator import (  # noqa: E501
    MultiStepLLMCommandGenerator,
)
from rasa.e2e_test.e2e_test_runner import E2ETestRunner
from rasa.llm_fine_tuning.annotation_module import annotate_e2e_tests
from rasa.llm_fine_tuning.llm_data_preparation_module import convert_to_fine_tuning_data
from rasa.llm_fine_tuning.paraphrasing.conversation_rephraser import (
    ConversationRephraser,
)
from rasa.llm_fine_tuning.paraphrasing_module import create_paraphrased_conversations
from rasa.llm_fine_tuning.storage import (
    FileStorageStrategy,
    StorageContext,
    StorageType,
)
from rasa.llm_fine_tuning.train_test_split_module import (
    CONVERSATIONAL_DATA_FORMAT,
    INSTRUCTION_DATA_FORMAT,
    split_llm_fine_tuning_data,
)
from rasa.shared.constants import (
    DEFAULT_MODELS_PATH,
    LLM_CONFIG_KEY,
)
from rasa.shared.utils.llm import (
    combine_custom_and_default_config,
    resolve_model_client_config,
)
from rasa.shared.utils.yaml import read_config_file
from rasa.utils.beta import ensure_beta_feature_is_enabled

DEFAULT_INPUT_E2E_TEST_PATH = "e2e_tests"
DEFAULT_OUTPUT_FOLDER = "output"
RESULT_SUMMARY_FILE = "result_summary.yaml"
PARAMETERS_FILE = "params.yaml"

structlogger = structlog.get_logger()


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add the llm fine-tuning subparser to `rasa test`.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    llm_parser = subparsers.add_parser(
        "llm",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Commands related to LLMs.",
    )
    llm_subparsers = llm_parser.add_subparsers()

    llm_finetune_parser = llm_subparsers.add_parser(
        "finetune",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Commands related to LLM fine-tuning.",
    )
    llm_finetune_subparser = llm_finetune_parser.add_subparsers()

    create_llm_finetune_data_preparation_subparser(llm_finetune_subparser, parents)


def create_llm_finetune_data_preparation_subparser(
    fine_tune_llm_parser: SubParsersAction,
    parents: List[argparse.ArgumentParser],
) -> argparse.ArgumentParser:
    """Create fine-tuning LLM data preparation subparser."""
    data_preparation_subparser = fine_tune_llm_parser.add_parser(
        "prepare-data",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Prepares data for LLM fine-tuning.",
    )

    data_preparation_subparser.set_defaults(func=prepare_llm_fine_tuning_data)

    add_data_preparation_arguments(data_preparation_subparser)
    add_model_param(data_preparation_subparser, add_positional_arg=False)
    add_endpoint_param(
        data_preparation_subparser,
        help_text="Configuration file for the model server and the connectors as a "
        "yml file.",
    )
    return data_preparation_subparser


def add_data_preparation_arguments(parser: argparse.ArgumentParser) -> None:
    """Arguments for preparing LLM fine-tuning data."""
    parser.add_argument(
        "-o",
        "--out",
        type=str,
        default=DEFAULT_OUTPUT_FOLDER,
        help="The output folder to store the data to.",
    )
    parser.add_argument(
        "path-to-e2e-test-cases",
        nargs="?",
        type=str,
        help="Input file or folder containing end-to-end test cases.",
        default=DEFAULT_INPUT_E2E_TEST_PATH,
    )

    add_sub_agents_param(parser)

    add_remote_storage_param(parser)

    rephrasing_arguments = parser.add_argument_group("Rephrasing Module")
    rephrasing_arguments.add_argument(
        "--num-rephrases",
        choices=range(0, 50),
        type=int,
        default=10,
        help="Number of rephrases to be generated per user utterance.",
    )
    rephrasing_arguments.add_argument(
        "--rephrase-config",
        type=str,
        default=None,
        help="Path to config file that contains the configuration of the "
        "rephrasing module.",
    )

    train_test_split_arguments = parser.add_argument_group("Train/Test Split Module")
    train_test_split_arguments.add_argument(
        "--train-frac",
        type=restricted_float,
        default=0.8,
        help="The amount of data that should go into the training dataset. The value "
        "should be >0.0 and <=1.0.",
    )
    train_test_split_arguments.add_argument(
        "--output-format",
        choices=[INSTRUCTION_DATA_FORMAT, CONVERSATIONAL_DATA_FORMAT],
        type=str,
        nargs="?",
        default=INSTRUCTION_DATA_FORMAT,
        help="Format of the output file.",
    )


def prepare_llm_fine_tuning_data(args: argparse.Namespace) -> None:
    """Prepare LLM fine-tuning data.

    Args:
        args: Commandline arguments.
    """
    ensure_beta_feature_is_enabled(
        "LLM fine-tuning recipe",
        env_flag=RASA_PRO_BETA_FINE_TUNING_RECIPE_ENV_VAR_NAME,
    )

    rephrase_config = (
        read_config_file(args.rephrase_config) if args.rephrase_config else {}
    )
    ConversationRephraser.validate_config(rephrase_config)

    # make sure the output directory exists
    output_dir = args.out
    rasa.shared.utils.io.create_directory(output_dir)

    # read e2e test cases
    path_to_test_cases = getattr(
        args, "path-to-e2e-test-cases", DEFAULT_INPUT_E2E_TEST_PATH
    )
    test_suite = read_test_cases(path_to_test_cases)
    # set up the e2e test runner
    e2e_test_runner = set_up_e2e_test_runner(args)

    if e2e_test_runner.agent.processor is None:
        rasa.shared.utils.cli.print_error(
            "No processor: Not able to retrieve flows and config from trained model."
        )
        sys.exit(0)

    flows = asyncio.run(e2e_test_runner.agent.processor.get_flows())
    _validate_llm_command_generator_present(e2e_test_runner)

    # set up storage context
    storage_context = create_storage_context(StorageType.FILE, output_dir)

    statistics = {}

    # 1. annotate e2e tests
    log_start_of_module("Annotation")
    conversations = annotate_e2e_tests(e2e_test_runner, test_suite, storage_context)
    statistics["num_input_e2e_tests"] = len(test_suite.test_cases)
    statistics["num_annotated_conversations"] = len(conversations)
    statistics["num_user_messages_across_conversations"] = sum(
        [len(conversation.get_user_messages()) for conversation in conversations]
    )
    statistics["num_user_messages_to_rephrase_across_conversations"] = sum(
        [
            len(conversation.get_user_messages_to_rephrase())
            for conversation in conversations
        ]
    )
    log_end_of_module("Annotation", statistics)

    # 2. paraphrase conversations
    log_start_of_module("Rephrasing")
    conversations, rephrase_config = asyncio.run(
        create_paraphrased_conversations(
            conversations,
            rephrase_config,
            args.num_rephrases,
            flows,
            e2e_test_runner.agent,
            storage_context,
        )
    )

    statistics["num_passing_rephrased_user_messages"] = sum(
        [conversation.get_number_of_rephrases(True) for conversation in conversations]
    )
    statistics["num_failing_rephrased_user_messages"] = sum(
        [conversation.get_number_of_rephrases(False) for conversation in conversations]
    )
    log_end_of_module("Rephrasing", statistics)

    # 3. create fine-tuning dataset
    log_start_of_module("LLM Data Preparation")
    llm_fine_tuning_data = asyncio.run(
        convert_to_fine_tuning_data(
            conversations, storage_context, e2e_test_runner.agent
        )
    )
    statistics["num_ft_data_points"] = len(llm_fine_tuning_data)
    log_end_of_module("LLM Data Preparation", statistics)

    # 4. create train/test split
    log_start_of_module("Train/Test Split")
    train_data, val_data = split_llm_fine_tuning_data(
        llm_fine_tuning_data,
        args.train_frac,
        args.output_format,
        storage_context,
        test_suite,
    )
    statistics["num_train_data_points"] = len(train_data)
    statistics["num_val_data_points"] = len(val_data)
    log_end_of_module("Train/Test Split", statistics)

    # write down params and statistics to a file
    write_params(args, rephrase_config, output_dir)
    write_statistics(statistics, output_dir)

    rasa.shared.utils.cli.print_success(
        f"Data and intermediate results are written to '{output_dir}'."
    )


def _get_llm_command_generator_config(e2e_test_runner: E2ETestRunner) -> Dict[str, Any]:
    train_schema = e2e_test_runner.agent.processor.model_metadata.train_schema  # type: ignore

    for node_name, node in train_schema.nodes.items():
        if node.matches_type(
            LLMBasedCommandGenerator, include_subtypes=True
        ) and not node.matches_type(
            MultiStepLLMCommandGenerator, include_subtypes=True
        ):
            # Configurations can reference model groups defined in the endpoints.yml
            resolved_llm_config = resolve_model_client_config(
                node.config.get(LLM_CONFIG_KEY, {}), node_name
            )
            llm_command_generator = cast(Type[LLMBasedCommandGenerator], node.uses)
            return combine_custom_and_default_config(
                resolved_llm_config, llm_command_generator.get_default_llm_config()
            )

    rasa.shared.utils.cli.print_error(
        "The provided model is not trained using 'SingleStepLLMCommandGenerator' or "
        "'CompactLLMCommandGenerator' or its subclasses. Without it, no data for "
        "fine-tuning can be generated. To resolve this, please include "
        "'SingleStepLLMCommandGenerator' or 'CompactLLMCommandGenerator' or its "
        "subclasses in your config and train your model."
    )
    sys.exit(1)


def _validate_llm_command_generator_present(
    e2e_test_runner: E2ETestRunner,
) -> None:
    train_schema = e2e_test_runner.agent.processor.model_metadata.train_schema  # type: ignore

    for _, node in train_schema.nodes.items():
        if node.matches_type(
            LLMBasedCommandGenerator, include_subtypes=True
        ) and not node.matches_type(
            MultiStepLLMCommandGenerator, include_subtypes=True
        ):
            return

    rasa.shared.utils.cli.print_error(
        "The provided model is not trained using 'SingleStepLLMCommandGenerator' or "
        "'CompactLLMCommandGenerator' or its subclasses. Without it, no data for "
        "fine-tuning can be generated. To resolve this, please include "
        "'SingleStepLLMCommandGenerator' or 'CompactLLMCommandGenerator' or its "
        "subclasses in your config and train your model."
    )
    sys.exit(1)


def log_start_of_module(module_name: str) -> None:
    log_info = f"Starting {module_name} Module"
    rasa.shared.utils.cli.print_info(
        f"{rasa.shared.utils.cli.pad(log_info, char='-')}\n"
    )


def log_end_of_module(module_name: str, statistics: Dict[str, int]) -> None:
    log_info = f"Finished {module_name} Module"
    rasa.shared.utils.cli.print_info(
        f"{rasa.shared.utils.cli.pad(log_info, char='-')}\n"
    )
    rasa.shared.utils.cli.print_color(
        "Current Statistics:", color=rasa.shared.utils.io.bcolors.BOLD
    )
    for key, value in statistics.items():
        rasa.shared.utils.cli.print_color(
            f"  {key}: {value}", color=rasa.shared.utils.io.bcolors.BOLD
        )


def restricted_float(x: Any) -> float:
    try:
        x = float(x)
    except ValueError:
        raise argparse.ArgumentTypeError("%r not a floating-point literal" % (x,))

    if x <= 0.0 or x > 1.0:
        raise argparse.ArgumentTypeError("%r not in range [0.0, 1.0]" % (x,))
    return x


def write_params(
    args: argparse.Namespace, rephrase_config: Dict[str, Any], output_path: str
) -> None:
    yaml_data = {
        "parameters": {
            "num_rephrases": args.num_rephrases,
            "rephrase_config": rephrase_config,
            "model": args.model,
            "endpoints": args.endpoints,
            "sub-agents": args.sub_agents,
            "remote-storage": args.remote_storage,
            "train_frac": args.train_frac,
            "output_format": args.output_format,
            "out": output_path,
        }
    }

    rasa.shared.utils.yaml.write_yaml(yaml_data, f"{output_path}/{PARAMETERS_FILE}")


def write_statistics(statistics: Dict[str, Any], output_path: str) -> None:
    rasa.shared.utils.yaml.write_yaml(
        statistics, f"{output_path}/{RESULT_SUMMARY_FILE}"
    )


def get_valid_endpoints(endpoints_file: str) -> AvailableEndpoints:
    endpoints = Configuration.initialise_endpoints(
        endpoints_path=Path(endpoints_file)
    ).endpoints

    # Ignore all endpoints apart from action server, model, nlu and nlg
    # to ensure InMemoryTrackerStore is being used instead of production
    # tracker store
    endpoints.tracker_store = None
    endpoints.lock_store = None
    endpoints.event_broker = None

    return endpoints


def set_up_e2e_test_runner(args: argparse.Namespace) -> E2ETestRunner:
    endpoints = get_valid_endpoints(args.endpoints)
    Configuration.initialise_sub_agents(args.sub_agents)

    if endpoints.model is None:
        args.model = validate_model_path(args.model, "model", DEFAULT_MODELS_PATH)

    try:
        return E2ETestRunner(
            remote_storage=args.remote_storage,
            model_path=args.model,
            model_server=endpoints.model,
            endpoints=endpoints,
            sub_agents_path=args.sub_agents,
        )
    except AgentNotReady as error:
        structlogger.error(
            "cli.finetune_llm.prepare_data.set_up_e2e_test_runner", error=error.message
        )
        sys.exit(1)


def create_storage_context(
    storage_type: StorageType, output_dir: str
) -> StorageContext:
    if storage_type == StorageType.FILE:
        strategy = FileStorageStrategy(output_dir)
    else:
        raise ValueError("Unsupported storage type")

    return StorageContext(strategy)



import os
from pathlib import Path

DEFAULT_CACHE_LOCATION = Path(".rasa", "cache")
DEFAULT_CACHE_NAME = "cache.db"
DEFAULT_CACHE_SIZE_MB = 1000

CACHE_LOCATION_ENV = "RASA_CACHE_DIRECTORY"
CACHE_DB_NAME_ENV = "RASA_CACHE_NAME"
CACHE_SIZE_ENV = "RASA_MAX_CACHE_SIZE"


def get_local_cache_location() -> Path:
    """Returns the location of the local cache."""
    return Path(os.environ.get(CACHE_LOCATION_ENV, DEFAULT_CACHE_LOCATION))


def get_max_cache_size() -> float:
    """Returns the maximum cache size."""
    return float(os.environ.get(CACHE_SIZE_ENV, DEFAULT_CACHE_SIZE_MB))


def get_cache_database_name() -> str:
    """Returns the database name in the cache."""
    return os.environ.get(CACHE_DB_NAME_ENV, DEFAULT_CACHE_NAME)



from __future__ import annotations

import asyncio
import copy
import datetime
import os
import queue
import time
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple

import structlog
from apscheduler.schedulers.background import BackgroundScheduler

import rasa.shared.core.trackers
from rasa.core.tracker_stores.tracker_store import TrackerStore
from rasa.privacy.constants import (
    TEXT_KEY,
    USER_CHAT_INACTIVITY_IN_MINUTES_ENV_VAR_NAME,
)
from rasa.privacy.event_broker_utils import create_event_brokers
from rasa.privacy.privacy_config import (
    PrivacyConfig,
    PrivacyPolicy,
    validate_sensitive_slots,
)
from rasa.privacy.privacy_filter import PrivacyFilter
from rasa.shared.core.events import Event, SlotSet, UserUttered, split_events
from rasa.shared.core.trackers import DialogueStateTracker, EventVerbosity

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop

    from rasa.core.brokers.broker import EventBroker
    from rasa.core.config.available_endpoints import AvailableEndpoints
    from rasa.shared.core.domain import Domain


structlogger = structlog.get_logger(__name__)


def wrap_async(func: Callable) -> Callable:
    """Wraps a function to be used as an async job in the background scheduler."""

    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return asyncio.run(func(*args, **kwargs))

    return wrapper


class BackgroundPrivacyManager:
    """Manages privacy-related tasks in the background.

    This class handles the anonymization and deletion of sensitive information
    in dialogue state trackers, as well as the streaming of anonymized events
    to event brokers. It uses background schedulers to periodically run these
    tasks and processes trackers from a queue to ensure that sensitive information
    is handled in a timely manner.
    """

    TRACKER_QUEUE_PROCESSING_TIMEOUT_IN_SECONDS = 2.0

    def __init__(
        self,
        endpoints: Optional["AvailableEndpoints"],
        event_loop: Optional["AbstractEventLoop"] = None,
        in_memory_tracker_store: Optional[TrackerStore] = None,
    ):
        self.config = (
            PrivacyConfig.from_dict(endpoints.privacy)
            if endpoints and endpoints.privacy
            else None
        )
        self.privacy_filter = (
            PrivacyFilter(self.config.anonymization_rules) if self.config else None
        )
        self.user_chat_inactivity_in_minutes = int(
            os.getenv(USER_CHAT_INACTIVITY_IN_MINUTES_ENV_VAR_NAME, 30)
        )

        if in_memory_tracker_store is not None:
            # if an in-memory tracker store is provided,
            # we need to keep the reference to it
            # so that the background jobs can access it.
            # We also set the event broker to None
            # to prevent it from publishing events
            # during the tracker store background jobs
            in_memory_tracker_store.event_broker = None
            tracker_store = in_memory_tracker_store
        else:
            # we recreate the tracker store here to ensure
            # that this instance has no event brokers
            # that could publish events during the tracker store
            # background jobs
            tracker_store = (
                TrackerStore.create(endpoints.tracker_store)
                if endpoints
                else TrackerStore.create(None)
            )

        self.tracker_store = tracker_store

        self.event_brokers: List["EventBroker"] = []
        self.event_loop = event_loop

        # Order of the initialisation is important
        # The tracker queue must be created before the scheduler
        # The can_consume_tracker_queue must be set to True before the scheduler starts
        self.tracker_queue: queue.Queue = queue.Queue()

        # This flag is used to stop the scheduler
        self.can_consume_from_tracker_queue = True
        self.background_scheduler = BackgroundScheduler()
        self.background_scheduler.add_job(
            self._consumer_queue, max_instances=1, id="event_broker_job"
        )

        self.previous_fire_time_deletion = datetime.datetime.now(
            tz=datetime.timezone.utc
        )
        self._configure_background_scheduler()
        self.background_scheduler.start()

    async def initialize(
        self, endpoints: Optional["AvailableEndpoints"]
    ) -> BackgroundPrivacyManager:
        """Initialize async attributes of the BackgroundPrivacyManager."""
        event_broker_endpoints = endpoints.event_broker if endpoints else None
        self.event_brokers = (
            await create_event_brokers(event_broker_endpoints, self.event_loop)
            if event_broker_endpoints
            else []
        )

        return self

    @classmethod
    async def create_instance(
        cls,
        endpoints: Optional["AvailableEndpoints"],
        event_loop: Optional["AbstractEventLoop"] = None,
        in_memory_tracker_store: Optional[TrackerStore] = None,
    ) -> BackgroundPrivacyManager:
        """Create an instance of BackgroundPrivacyManager."""
        instance = cls(endpoints, event_loop, in_memory_tracker_store)
        return await instance.initialize(endpoints)

    def stop(self) -> None:
        structlogger.debug("rasa.privacy_manager.stop_schedulers")
        self.can_consume_from_tracker_queue = False
        self.background_scheduler.shutdown(wait=False)

    def run(self, tracker: "DialogueStateTracker") -> None:
        self.tracker_queue.put(tracker)

    def process(
        self, tracker: "DialogueStateTracker", process_all: bool = False
    ) -> None:
        """Process the tracker to identify and anonymize sensitive information.

        Args:
            tracker: The tracker to process.
            process_all: If True, process all events in the tracker.
        """
        events = self.process_events(tracker, process_all=process_all)
        events_to_stream = events if events else tracker.events
        self.stream_events(events_to_stream, tracker.sender_id)

    def process_events(
        self, tracker: DialogueStateTracker, process_all: bool = False
    ) -> List[Event]:
        """Anonymize tracker events."""
        if (latest_message := self._get_latest_user_message(tracker)) is None:
            return []

        processed_events = list(tracker.events)
        prior_sensitive_slot_events: List[Event] = []

        if not process_all:
            additional_splitting_conditions = {
                TEXT_KEY: latest_message.text,
                "timestamp": latest_message.timestamp,
            }

            resulting_events = split_events(
                processed_events,
                UserUttered,
                additional_splitting_conditions=additional_splitting_conditions,
                include_splitting_event=True,
            )

            processed_events = resulting_events[1]
            prior_events = resulting_events[0]
            prior_tracker = DialogueStateTracker.from_events(
                sender_id=tracker.sender_id, evts=prior_events, user_id=tracker.user_id
            )
            prior_sensitive_slot_events = [
                event
                for event in prior_tracker.applied_events()
                if isinstance(event, SlotSet)
                and event.key in self.config.anonymization_rules  # type: ignore[union-attr]
            ]

        return self.privacy_filter.anonymize(  # type: ignore[union-attr]
            processed_events, prior_sensitive_slot_events
        )

    def stream_events(
        self,
        anonymized_events: List[Event],
        sender_id: str,
    ) -> None:
        """Stream anonymized events to the event broker."""
        if not self.event_brokers:
            structlogger.debug(
                "rasa.privacy_manager.no_event_broker_configured",
            )
            return None

        for event in anonymized_events:
            body = {"sender_id": sender_id}
            body.update(event.as_dict())
            for broker in self.event_brokers:
                broker.publish(body)

        return None

    def validate_sensitive_slots_in_domain(self, domain: "Domain") -> None:
        """Validate the sensitive slots defined in the privacy config against the domain."""  # noqa: E501
        if not self.config:
            structlogger.debug(
                "rasa.privacy_manager.no_sensitive_slots_configured",
            )
            return None

        # we need to set the domain in the tracker store
        # to prevent errors being raised about slots not found in the domain
        # during the background jobs
        self.tracker_store.domain = domain
        sensitive_slots = list(self.config.anonymization_rules.keys())
        return validate_sensitive_slots(sensitive_slots, domain)

    def _consumer_queue(self) -> None:
        while self.can_consume_from_tracker_queue:
            try:
                # Wait for 2 seconds for an event to be added to the queue
                # If no event is added to the queue, continue
                # This is done to avoid the scheduler to be stuck in the while loop
                # when we want to stop the scheduler
                tracker = self.tracker_queue.get(
                    timeout=self.TRACKER_QUEUE_PROCESSING_TIMEOUT_IN_SECONDS
                )
                self.process(tracker)
                self.tracker_queue.task_done()
            except queue.Empty:
                continue

    def _get_latest_user_message(
        self, tracker: DialogueStateTracker
    ) -> Optional[UserUttered]:
        """Check if a tracker should be processed."""
        if self.privacy_filter is None:
            structlogger.debug(
                "rasa.privacy_manager.no_privacy_rules_configured",
            )
            return None

        latest_user_message_event = tracker.get_last_event_for(
            UserUttered, event_verbosity=EventVerbosity.ALL
        )
        if latest_user_message_event is None or not isinstance(
            latest_user_message_event, UserUttered
        ):
            structlogger.debug(
                "rasa.privacy_manager.no_user_message.skipping_processing",
            )
            return None

        latest_user_message: UserUttered = latest_user_message_event
        if not latest_user_message.text:
            structlogger.debug(
                "rasa.privacy_manager.no_user_message.skipping_processing",
            )
            return None

        return latest_user_message

    @staticmethod
    def _has_session_been_anonymized(events: List[Event]) -> bool:
        """Check if the session has already been anonymized."""
        if not events:
            return False
        for event in reversed(events):
            if (
                hasattr(event, "anonymized_at")
                and getattr(event, "anonymized_at") is not None
            ):
                return True

        return False

    async def _run_tracker_store_anonymization(self) -> None:
        """Anonymize eligible tracker sessions in the tracker store."""
        structlogger.info(
            "rasa.privacy_manager.starting_tracker_store_anonymization",
            triggered_by="anonymization_cron_job",
        )

        keys = await self.tracker_store.keys()
        keys_copy = copy.deepcopy(list(keys))

        for key in keys_copy:
            full_tracker = await self.tracker_store.retrieve_full_tracker(key)

            if not full_tracker:
                structlogger.debug(
                    "rasa.privacy_manager.no_tracker_found_for_sender_id",
                    sender_id=key,
                )
                continue

            processed_events, already_anonymized_events, uneligible_events = (
                self._get_processed_events_after_anonymization(full_tracker)
            )

            if not processed_events:
                structlogger.debug(
                    "rasa.privacy_manager.no_events_to_anonymize_for_tracker",
                    sender_id=key,
                )
                continue

            all_events = (
                already_anonymized_events + processed_events + uneligible_events
            )
            updated_tracker = DialogueStateTracker.from_events(
                sender_id=key,
                evts=all_events,
                slots=full_tracker.slots.values(),
                user_id=full_tracker.user_id,
            )
            await self.tracker_store.delete(sender_id=key)
            await self.tracker_store.save(updated_tracker)

            structlogger.info(
                "rasa.privacy_manager.saved_tracker_after_anonymization",
                sender_id=key,
            )

    async def _run_tracker_store_deletion(self) -> None:
        """Delete eligible tracker sessions from the tracker store."""
        structlogger.info(
            "rasa.privacy_manager.starting_tracker_store_deletion",
            triggered_by="deletion_cron_job",
        )
        keys = await self.tracker_store.keys()

        # Make a copy of the keys to avoid modifying the list while iterating
        keys_copy = copy.deepcopy(list(keys))

        for key in keys_copy:
            full_tracker = await self.tracker_store.retrieve_full_tracker(key)

            if not full_tracker:
                structlogger.debug(
                    "rasa.privacy_manager.no_tracker_found_for_sender_id",
                    sender_id=key,
                )
                continue

            events_to_be_retained = self._get_events_to_be_retained_after_deletion(
                full_tracker
            )

            if not events_to_be_retained:
                await self.tracker_store.delete(sender_id=key)
                structlogger.info(
                    "rasa.privacy_manager.tracker_session_deleted",
                    sender_id=full_tracker.sender_id,
                    triggered_by="deletion_cron_job",
                )
                continue

            tracker = DialogueStateTracker.from_events(
                sender_id=key,
                evts=events_to_be_retained,
                slots=full_tracker.slots.values(),
                user_id=full_tracker.user_id,
            )
            await self.tracker_store.update(tracker)

            structlogger.info(
                "rasa.privacy_manager.overwritten_tracker",
                sender_id=key,
                event_info="Deleted eligible events and saved "
                "tracker with events not scheduled "
                "for deletion yet.",
            )

    async def _run_tracker_store_background_jobs_sequentially(self) -> None:
        """Run the tracker store background jobs.

        If both anonymization and deletion policies are configured,
        we need to ensure that the background job timings do not
        overlap to prevent race conditions when accessing the
        tracker store.

        The scheduler will run the anonymization job first,
        and then the deletion job if the current time is past
        the next scheduled time for deletion.
        """
        await self._run_tracker_store_anonymization()

        now = datetime.datetime.now(tz=datetime.timezone.utc)
        next_fire_time = (
            self.config.tracker_store_settings.deletion_policy.cron.get_next_fire_time(  # type: ignore[union-attr]
                self.previous_fire_time_deletion,
                now=now,
            )
        )

        if next_fire_time and now >= next_fire_time:
            await self._run_tracker_store_deletion()
            self.previous_fire_time_deletion = next_fire_time

        return None

    def _add_anonymization_job(self) -> None:
        wrapped_anonymization = wrap_async(self._run_tracker_store_anonymization)
        self.background_scheduler.add_job(
            wrapped_anonymization,
            trigger=self.config.tracker_store_settings.anonymization_policy.cron,  # type: ignore[union-attr]
            max_instances=1,
            id="anonymization_cron_job",
        )

    def _add_deletion_job(self) -> None:
        wrapped_deletion = wrap_async(self._run_tracker_store_deletion)
        self.background_scheduler.add_job(
            wrapped_deletion,
            trigger=self.config.tracker_store_settings.deletion_policy.cron,  # type: ignore[union-attr]
            max_instances=1,
            id="deletion_cron_job",
        )

    def _add_sequential_job(self) -> None:
        sequential_dispatcher = wrap_async(
            self._run_tracker_store_background_jobs_sequentially
        )
        self.background_scheduler.add_job(
            sequential_dispatcher,
            trigger=self.config.tracker_store_settings.anonymization_policy.cron,  # type: ignore[union-attr]
            max_instances=1,
            id="anonymization_and_deletion_cron_job",
        )

    def _configure_background_scheduler(self) -> None:
        """Configure the background scheduler."""
        tracker_store_settings_configured = (
            self.config is not None and self.config.tracker_store_settings is not None
        )
        anonymization_policy = (
            self.config.tracker_store_settings.anonymization_policy  # type: ignore[union-attr]
            if tracker_store_settings_configured
            else None
        )
        deletion_policy = (
            self.config.tracker_store_settings.deletion_policy  # type: ignore[union-attr]
            if tracker_store_settings_configured
            else None
        )

        if (
            tracker_store_settings_configured
            and anonymization_policy is not None
            and deletion_policy is not None
        ):
            next_fire_time_anonymization = get_next_fire_time(anonymization_policy)
            next_fire_time_deletion = get_next_fire_time(deletion_policy)

            # If both anonymization and deletion policies are configured
            # to start on the same date,
            # we need to run them sequentially to avoid race conditions
            if (
                next_fire_time_anonymization is not None
                and next_fire_time_deletion is not None
                and next_fire_time_anonymization.date()
                == next_fire_time_deletion.date()
            ):
                self._add_sequential_job()
            else:
                self._add_anonymization_job()
                self._add_deletion_job()

        elif tracker_store_settings_configured and anonymization_policy is not None:
            self._add_anonymization_job()

        elif tracker_store_settings_configured and deletion_policy is not None:
            self._add_deletion_job()

    def _get_processed_events_after_anonymization(
        self,
        full_tracker: DialogueStateTracker,
    ) -> Tuple[List[Event], List[Event], List[Event]]:
        """Get processed events after anonymization job."""
        multiple_tracker_sessions = (
            rasa.shared.core.trackers.get_trackers_for_conversation_sessions(
                full_tracker
            )
        )

        processed_events = []
        already_anonymized_events = []
        uneligible_events = []

        for session in multiple_tracker_sessions:
            has_session_been_anonymized = self._has_session_been_anonymized(
                list(session.events)
            )

            if has_session_been_anonymized:
                structlogger.debug(
                    "rasa.privacy_manager.session_already_anonymized",
                    session_id=session.sender_id,
                )
                already_anonymized_events.extend(list(session.events))
                continue

            current_time = time.time()

            last_event_timestamp = (
                str(datetime.datetime.fromtimestamp(session.events[-1].timestamp))
                if session.events
                else "N/A"
            )

            if session.events and current_time - session.events[-1].timestamp > (
                self.user_chat_inactivity_in_minutes * 60
                + self.config.tracker_store_settings.anonymization_policy.min_after_session_end  # type: ignore[union-attr] # noqa: E501
                * 60
            ):
                structlogger.info(
                    "rasa.privacy_manager.anonymizing_tracker_session",
                    sender_id=session.sender_id,
                    last_event_timestamp=last_event_timestamp,
                    triggered_by="anonymization_cron_job",
                )
                events = self.process_events(session, process_all=True)
                processed_events.extend(events)
            else:
                # If the session is not valid for anonymization,
                # we still want to write them back to the tracker store
                events = list(session.events)
                uneligible_events.extend(events)
                structlogger.debug(
                    "rasa.privacy_manager.session_not_valid_for_anonymization",
                    sender_id=session.sender_id,
                    session_id=session.sender_id,
                    last_event_timestamp=last_event_timestamp,
                )
        return processed_events, already_anonymized_events, uneligible_events

    def _get_events_to_be_retained_after_deletion(
        self, full_tracker: DialogueStateTracker
    ) -> List[Event]:
        """Get the events to be retained after deletion."""
        multiple_tracker_sessions = (
            rasa.shared.core.trackers.get_trackers_for_conversation_sessions(
                full_tracker
            )
        )
        events_to_be_retained: List[Event] = []
        for session in multiple_tracker_sessions:
            current_time = time.time()
            if session.events and (
                current_time - session.events[-1].timestamp
                <= (
                    self.user_chat_inactivity_in_minutes * 60
                    + self.config.tracker_store_settings.deletion_policy.min_after_session_end  # type: ignore[union-attr] # noqa: E501
                    * 60
                )
            ):
                events_to_be_retained.extend(session.events)
            else:
                last_event_timestamp = (
                    str(datetime.datetime.fromtimestamp(session.events[-1].timestamp))
                    if session.events
                    else "N/A"
                )

                structlogger.info(
                    "rasa.privacy_manager.tracker_session_scheduled_for_deletion",
                    sender_id=full_tracker.sender_id,
                    last_event_timestamp=last_event_timestamp,
                    triggered_by="deletion_cron_job",
                )

        return events_to_be_retained


def get_next_fire_time(
    privacy_policy: PrivacyPolicy,
) -> Optional[datetime.datetime]:
    """Get the next fire time for the privacy policy."""
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    return privacy_policy.cron.get_next_fire_time(None, now=now)



import argparse
import logging
import uuid
from pathlib import Path
from typing import List

from rasa import telemetry
from rasa.cli import SubParsersAction
from rasa.cli.arguments import shell as arguments
from rasa.cli.arguments.default_arguments import add_sub_agents_param
from rasa.core.config.configuration import Configuration
from rasa.engine.storage.local_model_storage import LocalModelStorage
from rasa.exceptions import ModelNotFound
from rasa.model import get_local_model
from rasa.shared.constants import ASSISTANT_ID_KEY
from rasa.shared.data import TrainingType
from rasa.shared.utils.cli import print_error

logger = logging.getLogger(__name__)


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add all shell parsers.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    shell_parser = subparsers.add_parser(
        "shell",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help=(
            "Loads your trained model and lets you talk to your "
            "assistant on the command line."
        ),
    )
    shell_parser.set_defaults(func=shell)

    shell_parser.add_argument(
        "--conversation-id",
        default=uuid.uuid4().hex,
        required=False,
        help="Set the conversation ID.",
    )

    add_sub_agents_param(shell_parser)

    run_subparsers = shell_parser.add_subparsers()

    shell_nlu_subparser = run_subparsers.add_parser(
        "nlu",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Interprets messages on the command line using your NLU model.",
    )

    shell_nlu_subparser.set_defaults(func=shell_nlu)

    arguments.set_shell_arguments(shell_parser)
    arguments.set_shell_nlu_arguments(shell_nlu_subparser)


def shell_nlu(args: argparse.Namespace) -> None:
    """Talk with an NLU only bot though the command line."""
    import rasa.nlu.run
    from rasa.cli.validation.config_path_validation import get_validated_path
    from rasa.shared.constants import DEFAULT_MODELS_PATH

    args.connector = "cmdline"

    model_path = get_validated_path(args.model, "model", DEFAULT_MODELS_PATH)

    try:
        model = get_local_model(model_path)
    except ModelNotFound:
        print_error(
            "No model found. Train a model before running the "
            "server using `rasa train nlu`."
        )
        return

    metadata = LocalModelStorage.metadata_from_archive(model)

    if metadata.assistant_id is None:
        print_error(
            f"The model metadata does not contain a value for the '{ASSISTANT_ID_KEY}' "
            f"attribute. Check that 'config.yml' file contains a value for "
            f"the '{ASSISTANT_ID_KEY}' key and re-train the model.",
        )

    if metadata.training_type == TrainingType.CORE:
        print_error(
            "No NLU model found. Train a model before running the "
            "server using `rasa train nlu`."
        )
        return

    telemetry.track_shell_started("nlu", metadata.assistant_id)
    rasa.nlu.run.run_cmdline(model)


def shell(args: argparse.Namespace) -> None:
    """Talk with a bot though the command line."""
    from rasa.cli.validation.config_path_validation import get_validated_path
    from rasa.shared.constants import DEFAULT_MODELS_PATH

    args.connector = "cmdline"
    # Load endpoints with proper endpoint file location
    # This will initialise the endpoints singleton properly so that
    # it can be used safely throughout the codebase with
    # `Configuration.get_instance().endpoints`
    Configuration.initialise_endpoints(endpoints_path=Path(args.endpoints))
    Configuration.initialise_sub_agents(args.sub_agents)
    model = get_validated_path(args.model, "model", DEFAULT_MODELS_PATH)

    try:
        model = get_local_model(model)
    except ModelNotFound:
        print_error(
            "No model found. Train a model before running the "
            "server using `rasa train`."
        )
        return

    metadata = LocalModelStorage.metadata_from_archive(model)

    if metadata.assistant_id is None:
        print_error(
            f"The model metadata does not contain a value for the '{ASSISTANT_ID_KEY}' "
            f"attribute. Check that 'config.yml' file contains a value for "
            f"the '{ASSISTANT_ID_KEY}' key and re-train the model.",
        )

    if metadata.training_type == TrainingType.NLU:
        import rasa.nlu.run

        telemetry.track_shell_started("nlu", metadata.assistant_id)

        rasa.nlu.run.run_cmdline(model)
    else:
        import rasa.cli.run

        telemetry.track_shell_started("rasa", metadata.assistant_id)

        rasa.cli.run.run(args)



"""Agent hooks for tracking tool execution events."""

import asyncio
from typing import Any

import structlog
from agents import Agent, AgentHooks, Tool
from agents.run_context import RunContextWrapper

from rasa.builder.copilot.models import MCPToolCall, MCPToolCallStatus

structlogger = structlog.get_logger()


class RasaCopilotHooks(AgentHooks):
    """Custom AgentHooks for tracking MCP tool events.

    This hooks class captures tool start and end events and queues them
    for processing by the response handler.
    """

    def __init__(self, mcp_queue: asyncio.Queue[MCPToolCall]) -> None:
        """Initialize the hooks with a queue for MCP tool call events.

        Args:
            mcp_queue: Queue to put MCPToolCall events for processing.
        """
        self._mcp_queue = mcp_queue

    async def on_tool_start(
        self, context: RunContextWrapper, agent: Agent, tool: Tool
    ) -> None:
        """Called when a tool starts execution."""
        structlogger.debug(
            "agent_sdk.tool.start",
            tool_name=tool.name,
        )
        # Queue MCP tool call event
        await self._mcp_queue.put(
            MCPToolCall(
                tool_name=tool.name,
                status=MCPToolCallStatus.CALLED,
            )
        )

    async def on_tool_end(
        self, context: RunContextWrapper, agent: Agent, tool: Tool, result: Any
    ) -> None:
        """Called when a tool completes execution."""
        structlogger.debug(
            "agent_sdk.tool.end",
            tool_name=tool.name,
        )
        # Queue MCP tool completion event
        await self._mcp_queue.put(
            MCPToolCall(
                tool_name=tool.name,
                status=MCPToolCallStatus.COMPLETED,
                output=result,
            )
        )



"""This module regroups custom validation functions, and it is
loaded as an extension of the pykwalify library:

https://pykwalify.readthedocs.io/en/latest/extensions.html#extensions
"""

from typing import Any, Dict, List, Text, Union

from pykwalify.errors import SchemaError


def require_response_keys(
    responses: List[Dict[Text, Any]], _: Dict, __: Text
) -> Union[SchemaError, bool]:
    """Validates that response dicts have either the "text" key or the "custom" key."""
    for response in responses:
        if not isinstance(response, dict):
            # this is handled by other validation rules
            continue

        if response.get("text") is None and not response.get("custom"):
            return SchemaError(
                "Missing 'text' or 'custom' key in response or "
                "null 'text' value in response."
            )

        conditions = response.get("condition", [])
        if isinstance(conditions, str):
            continue

        for condition in conditions:
            if not isinstance(condition, dict):
                return SchemaError("Condition must be a dictionary.")
            if not all(key in condition for key in ("type", "name", "value")):
                return SchemaError(
                    "Condition must have 'type', 'name', and 'value' keys."
                )

            if condition.get("type") != "slot":
                return SchemaError("Condition type must be of type `slot`.")

    return True



from rasa.shared.exceptions import RasaException


class InvalidLangfuseConfigException(RasaException):
    """Raised when a Langfuse config is invalid."""

    def __init__(self, message: str) -> None:
        """Initializes the exception."""
        self.message = message

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return self.message


class DuplicateTracingConfigException(RasaException):
    """Raised when a duplicate tracing config is found."""

    def __init__(self, message: str) -> None:
        """Initializes the exception."""
        self.message = message

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return self.message



from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Coroutine

from rasa.builder import config
from rasa.builder.telemetry.langfuse.langfuse_compat import (
    is_langfuse_available,
    langfuse,
)
from rasa.builder.telemetry.langfuse.shared import (
    update_generation_span_with_usage_statistics,
)

if TYPE_CHECKING:
    from rasa.builder.copilot.message_classifier.message_classifier import (
        MessageClassifier,
    )
    from rasa.builder.copilot.message_classifier.models import MessageClassifierResult
    from rasa.builder.copilot.models import CopilotContext


class MessageClassifierLangfuseTelemetry:
    """Telemetry for MessageClassifier LLM classification with Langfuse."""

    @staticmethod
    def trace_classification(
        func: Callable[..., Coroutine[Any, Any, "MessageClassifierResult"]],
    ) -> Callable[..., Coroutine[Any, Any, "MessageClassifierResult"]]:
        """Decorator for tracing MessageClassifier.classify() LLM calls.

        This decorator handles Langfuse tracing for the classification LLM call
        by creating a generation observation within the current trace.

        Args:
            func: The async classify method to wrap.

        Returns:
            Wrapped function with Langfuse telemetry.
        """
        if not is_langfuse_available():
            return func

        @wraps(func)
        async def wrapper(
            self: "MessageClassifier", context: "CopilotContext"
        ) -> "MessageClassifierResult":
            langfuse_client = langfuse.get_client()

            # Extract the last user message for logging
            last_user_message = context.get_last_user_message()
            user_message_text = ""
            if last_user_message:
                # Extract text from content blocks
                for content in last_user_message.content:
                    if hasattr(content, "text"):
                        user_message_text = content.text
                        break

            with langfuse_client.start_as_current_generation(
                name=f"{self.__class__.__name__}.{func.__name__}",
                input={"user_message": user_message_text},
            ) as generation:
                # Call the original classification function
                result = await func(self, context)

                # Import here to avoid circular dependency
                from rasa.builder.copilot.message_classifier.message_classifier import (
                    MessageClassifier,
                )

                # Update the span with classification result and model parameters
                generation.update(
                    model_parameters={
                        "model": config.ORCHESTRATOR_MODEL,
                        "temperature": MessageClassifier.CLASSIFICATION_TEMPERATURE,
                        "max_tokens": MessageClassifier.CLASSIFICATION_MAX_TOKENS,
                    },
                    output={
                        "raw_response": result.raw_response,
                        "category": result.category.value,
                        "requires_full_copilot": result.requires_full_copilot,
                    },
                )

                # Update the span's usage statistics
                if result.classification_usage:
                    update_generation_span_with_usage_statistics(
                        generation, result.classification_usage
                    )

                return result

        return wrapper


class MessageClassifierResponseHandlerLangfuseTelemetry:
    """Telemetry for MessageClassifierResponseHandler LLM generation with Langfuse."""

    @staticmethod
    def trace_response_generation(
        generation_type: str, max_tokens: int
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """Decorator factory for tracing greeting/goodbye generation LLM calls.

        This decorator handles Langfuse tracing for response generation
        by creating a generation observation within the current trace.

        Args:
            generation_type: Type of generation ("greeting" or "goodbye").
            max_tokens: Maximum tokens used in the LLM call.

        Returns:
            Decorator function that wraps the streaming method.
        """

        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            if not is_langfuse_available():
                return func

            @wraps(func)
            async def wrapper(self: Any) -> Any:
                langfuse_client = langfuse.get_client()

                with langfuse_client.start_as_current_generation(
                    name=f"MessageClassifierResponseHandler.generate_{generation_type}",
                    input={"user_message": self._user_message},
                ) as generation:
                    output_tokens: list[str] = []

                    # Call the original streaming function and capture output
                    async for token in func(self):
                        if hasattr(token, "content"):
                            output_tokens.append(token.content)
                        yield token

                    # Import here to avoid circular dependency
                    from rasa.builder.copilot.response_handling import (
                        message_classifier_response_handler as handler_module,
                    )

                    # Update the span with model parameters and output
                    handler_class = handler_module.MessageClassifierResponseHandler
                    generation.update(
                        model_parameters={
                            "model": config.ORCHESTRATOR_MODEL,
                            "temperature": handler_class.GENERATION_TEMPERATURE,
                            "max_tokens": max_tokens,
                            "stream": True,
                        },
                        output="".join(output_tokens),
                    )

                    # Update the span's usage statistics
                    if self._generation_usage:
                        update_generation_span_with_usage_statistics(
                            generation, self._generation_usage
                        )

            return wrapper

        return decorator



"""Pydantic models for MCP server input/output validation."""

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, validator

# Common field description constant
ERROR_FIELD_DESCRIPTION = "Error message if failed"


class FileUpdate(BaseModel):
    """Model for a single file update operation."""

    path: str = Field(
        ...,
        description="Relative path to the file within the project",
        min_length=1,
        max_length=255,
    )
    content: str = Field(..., description="Complete content to write to the file")

    @validator("path")
    def validate_path(cls, v: str) -> str:
        """Validate file path for security concerns."""
        # Prevent path traversal
        if ".." in v:
            raise ValueError("Path traversal is not allowed (contains '..')")

        # Prevent absolute paths
        if v.startswith("/"):
            raise ValueError("Absolute paths are not allowed")

        # Prevent hidden files
        if v.startswith(".") or "/." in v:
            raise ValueError("Cannot write to hidden files or directories")

        # Basic character validation
        allowed_chars = set(
            "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-./"
        )
        if not all(c in allowed_chars for c in v):
            raise ValueError("Path contains invalid characters")

        return v


class MultiFileUpdate(BaseModel):
    """Model for updating multiple files at once."""

    files: Dict[str, str] = Field(
        ...,
        description="Dictionary mapping file paths to their new contents",
    )

    @validator("files")
    def validate_files(cls, v: Dict[str, str]) -> Dict[str, str]:
        """Validate all file paths in the batch update."""
        # Check minimum and maximum number of files
        if len(v) < 1:
            raise ValueError("At least one file must be provided")
        if len(v) > 50:
            raise ValueError("Cannot update more than 50 files at once")

        # Validate each file path
        for path in v.keys():
            # Reuse the path validation from FileUpdate
            FileUpdate(path=path, content=v[path])
        return v


class SearchQuery(BaseModel):
    """Model for documentation search queries."""

    query: str = Field(
        ...,
        description="The search query to find relevant documentation",
        min_length=2,
        max_length=500,
    )

    @validator("query")
    def validate_query(cls, v: str) -> str:
        """Validate and sanitize search query."""
        # Trim whitespace
        v = v.strip()

        if len(v) < 2:
            raise ValueError("Query must be at least 2 characters long")

        return v


class FilePathInput(BaseModel):
    """Model for file path inputs."""

    file_path: str = Field(
        ...,
        description="Relative path to the file within the project",
        min_length=1,
        max_length=255,
    )

    @validator("file_path")
    def validate_file_path(cls, v: str) -> str:
        """Validate file path."""
        # Prevent path traversal
        if ".." in v:
            raise ValueError("Path traversal is not allowed")

        # Prevent absolute paths
        if v.startswith("/"):
            raise ValueError("Absolute paths are not allowed")

        return v


# =============================================================================
# OUTPUT MODELS - Structured outputs for MCP tools
# =============================================================================


class DocumentSearchResult(BaseModel):
    """A single document from the search results."""

    index: int = Field(description="Position in the search results")
    title: str = Field(description="Document title")
    url: str = Field(description="Document URL")
    content: str = Field(description="Document content snippet")


class DocumentSearchResponse(BaseModel):
    """Response from documentation search."""

    documents: List[DocumentSearchResult] = Field(
        default_factory=list, description="List of matching documents"
    )
    error: Optional[str] = Field(
        default=None, description="Error message if search failed"
    )


class FileListResponse(BaseModel):
    """Response from listing project files."""

    success: bool = Field(description="Whether the operation succeeded")
    tree: Optional[str] = Field(default=None, description="Visual tree representation")
    files: List[str] = Field(default_factory=list, description="List of file paths")
    count: int = Field(default=0, description="Number of files")
    directories: List[str] = Field(
        default_factory=list, description="List of directories"
    )
    error: Optional[str] = Field(default=None, description=ERROR_FIELD_DESCRIPTION)


class FileContentResponse(BaseModel):
    """Response from reading a single file."""

    file_path: Optional[str] = Field(default=None, description="Path to the file")
    content: Optional[str] = Field(default=None, description="File content")
    exists: bool = Field(default=True, description="Whether the file exists")
    error: Optional[str] = Field(default=None, description=ERROR_FIELD_DESCRIPTION)


class ReadFilesResponse(BaseModel):
    """Response from reading multiple project files."""

    files: Dict[str, Optional[str]] = Field(
        default_factory=dict, description="Map of file paths to contents"
    )
    count: int = Field(default=0, description="Number of files read")
    error: Optional[str] = Field(default=None, description=ERROR_FIELD_DESCRIPTION)


class WriteFileResponse(BaseModel):
    """Response from writing a file."""

    success: bool = Field(description="Whether the write succeeded")
    file_path: str = Field(description="Path to the file")
    message: str = Field(description="Status message")


class FileUpdateFailure(BaseModel):
    """Details about a failed file update."""

    file_path: Optional[str] = Field(default=None, description="Path that failed")
    error: str = Field(description="Error message")


class UpdateFilesResponse(BaseModel):
    """Response from updating multiple files."""

    success: bool = Field(description="Whether all updates succeeded")
    updated: List[str] = Field(default_factory=list, description="Files updated")
    failed: List[FileUpdateFailure] = Field(
        default_factory=list, description="Files that failed"
    )
    message: str = Field(description="Status message")


class ValidationErrorDetail(BaseModel):
    """A single validation error."""

    level: Optional[str] = Field(default="error", description="Error severity")
    message: str = Field(description="Error message")
    details: Optional[Dict[str, Any]] = Field(
        default=None, description="Additional error details"
    )


class ValidationResponse(BaseModel):
    """Response from project validation."""

    success: bool = Field(description="Whether validation passed")
    errors: Optional[List[ValidationErrorDetail]] = Field(
        default=None, description="List of validation errors"
    )
    message: str = Field(description="Validation status message")


class TrainingResponse(BaseModel):
    """Response from model training."""

    success: bool = Field(description="Whether training succeeded")
    model_path: Optional[str] = Field(default=None, description="Path to trained model")
    message: str = Field(description="Training status message")
    agent_reloaded: Optional[bool] = Field(
        default=None, description="Whether the agent was reloaded in the server"
    )


class BotResponse(BaseModel):
    """A single response from the bot."""

    text: Optional[str] = Field(default=None, description="Response text")
    image: Optional[str] = Field(default=None, description="Image URL")
    buttons: Optional[List[Dict[str, Any]]] = Field(
        default=None, description="Button options"
    )
    custom: Optional[Dict[str, Any]] = Field(default=None, description="Custom payload")


class ConversationTurn(BaseModel):
    """A single turn in the conversation."""

    user_message: str = Field(description="The user's message")
    bot_responses: List[Dict[str, Any]] = Field(
        default_factory=list, description="Bot responses to the message"
    )


class TrackerContextOutput(BaseModel):
    """Tracker context from the conversation."""

    conversation_turns: List[Dict[str, Any]] = Field(
        default_factory=list, description="Conversation history"
    )
    current_state: Dict[str, Any] = Field(
        default_factory=dict, description="Current conversation state"
    )


class TalkToAssistantResponse(BaseModel):
    """Response from talking to the assistant."""

    success: bool = Field(description="Whether the conversation succeeded")
    session_id: str = Field(description="Unique session ID for this conversation")
    message_count: int = Field(description="Number of messages sent")
    conversation: List[ConversationTurn] = Field(
        default_factory=list, description="The conversation history"
    )
    tracker_context: Optional[TrackerContextOutput] = Field(
        default=None, description="Full tracker context after conversation"
    )
    error: Optional[str] = Field(default=None, description=ERROR_FIELD_DESCRIPTION)



"""Service for handling LLM interactions."""

from typing import Any, Dict, Optional

import openai
import structlog

from rasa.builder import config
from rasa.builder.copilot.history_store import (
    CopilotHistoryStore,
    SQLiteCopilotHistoryStore,
)
from rasa.builder.guardrails.clients import (
    GuardrailsClient,
    LakeraAIGuardrails,
)
from rasa.builder.guardrails.policy_checker import GuardrailsPolicyChecker

structlogger = structlog.get_logger()


class LLMService:
    """Handles OpenAI LLM interactions with caching for efficiency."""

    def __init__(self) -> None:
        self._client: Optional[openai.AsyncOpenAI] = None
        self._domain_schema: Optional[Dict[str, Any]] = None
        self._flows_schema: Optional[Dict[str, Any]] = None
        self._guardrails: Optional[GuardrailsClient] = None
        self._guardrails_policy_checker: Optional[GuardrailsPolicyChecker] = None
        self._history_store: Optional[CopilotHistoryStore] = None

    @property
    def guardrails(self) -> Optional[GuardrailsClient]:
        """Get or lazy create guardrails instance."""
        if not config.ENABLE_GUARDRAILS:
            return None
        # TODO: Replace with Open Source guardrails implementation once it's ready
        try:
            if self._guardrails is None:
                self._guardrails = LakeraAIGuardrails()
            return self._guardrails
        except Exception as e:
            structlogger.error(
                "llm_service.guardrails.error",
                event_info="LLM Service: Error getting guardrails instance.",
                error=str(e),
            )
            raise

    @property
    def guardrails_policy_checker(self) -> Optional[GuardrailsPolicyChecker]:
        """Get or lazy create guardrails policy checker instance."""
        try:
            if self._guardrails_policy_checker is None and self.guardrails is not None:
                self._guardrails_policy_checker = GuardrailsPolicyChecker(
                    self.guardrails
                )
            return self._guardrails_policy_checker
        except Exception as e:
            structlogger.error(
                "llm_service.guardrails_policy_checker.error",
                event_info=(
                    "LLM Service: Error getting guardrails policy checker instance."
                ),
                error=str(e),
            )
            raise

    @property
    def history_store(self) -> CopilotHistoryStore:
        """Get or lazy create history store instance."""
        if self._history_store is None:
            database_path = config.COPILOT_HISTORY_SQLITE_PATH
            structlogger.info("llm_service.history_store.backend", path=database_path)
            self._history_store = SQLiteCopilotHistoryStore(database_path)

        try:
            return self._history_store
        except Exception as e:
            structlogger.error(
                "llm_service.history_store.error",
                event_info="LLM Service: Error getting history store instance.",
                error=str(e),
            )
            raise


# Global service instance
llm_service = LLMService()



"""Centralized langfuse import handling.

All langfuse imports MUST go through this module. Direct imports of langfuse
are banned via ruff (see pyproject.toml banned-api configuration).

This provides:
- Graceful fallback when langfuse is not installed (optional dependency)
- No-op implementations for decorators like @observe
- Context manager that only executes block when langfuse is available

Usage:
    from rasa.builder.telemetry.langfuse.langfuse_compat import observe, with_langfuse

    # For decorators (works whether langfuse is installed or not):
    @observe()
    def my_function(): ...

    # For conditional code that requires langfuse:
    with with_langfuse() as lf:
        if lf:  # Block only runs if langfuse is available
            client = lf.get_client()
"""

from contextlib import contextmanager
from functools import wraps
from types import ModuleType
from typing import Any, Callable, Generator, Optional, TypeVar

F = TypeVar("F", bound=Callable[..., Any])


def _no_op_observe(*args: Any, **kwargs: Any) -> Callable[[F], F]:
    """No-op implementation of langfuse.observe decorator that preserves types."""

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*f_args: Any, **f_kwargs: Any) -> Any:
            return func(*f_args, **f_kwargs)

        return wrapper  # type: ignore[return-value]

    return decorator


try:
    import langfuse as _langfuse  # noqa: TID251

    _LANGFUSE_AVAILABLE = True
    langfuse: Any = _langfuse
except ImportError:
    _LANGFUSE_AVAILABLE = False

    class _MockLangfuse:
        """Mock langfuse module when the real one isn't installed."""

        observe = staticmethod(_no_op_observe)

        def get_client(self) -> None:
            return None

    langfuse = _MockLangfuse()


@contextmanager
def with_langfuse() -> Generator[Optional[ModuleType], None, None]:
    """Context manager that yields the langfuse module if available.

    If langfuse is not installed, yields None and the block still executes
    but the caller must check the yielded value.

    Usage:
        with with_langfuse() as lf:
            if lf:
                lf.get_client().update_current_trace(...)
    """
    if _LANGFUSE_AVAILABLE:
        yield _langfuse
    else:
        yield None


def is_langfuse_available() -> bool:
    """Check if langfuse is available.

    Use this for decoration-time checks where context manager isn't suitable.
    """
    return _LANGFUSE_AVAILABLE


def require_langfuse() -> ModuleType:
    """Return the langfuse module or raise ImportError if not available.

    Use this in modules that absolutely require langfuse (e.g., evaluators).

    Raises:
        ImportError: If langfuse is not installed.
    """
    if not _LANGFUSE_AVAILABLE:
        raise ImportError(
            "langfuse is required for this functionality. "
            "Install it with: pip install rasa-pro[monitoring]"
        )
    return _langfuse


# Re-export for convenience - use the typed no-op when langfuse unavailable
def observe(*args: Any, **kwargs: Any) -> Callable[[F], F]:
    """Typed observe decorator that works whether langfuse is available or not."""
    return langfuse.observe(*args, **kwargs)



import typing
from typing import Any, Dict, List, Optional, Text, Type

# importing spacy at module level to ensure error is raised if spacy is not installed
import spacy

from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.nlu.constants import SPACY_DOCS
from rasa.nlu.tokenizers.tokenizer import Token, Tokenizer
from rasa.nlu.utils.spacy_utils import SpacyNLP
from rasa.shared.nlu.training_data.message import Message

if typing.TYPE_CHECKING:
    from spacy.tokens.doc import Doc

POS_TAG_KEY = "pos"


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.MESSAGE_TOKENIZER, is_trainable=False
)
class SpacyTokenizer(Tokenizer):
    """Tokenizer that uses SpaCy."""

    @classmethod
    def required_components(cls) -> List[Type]:
        """Components that should be included in the pipeline before this component."""
        return [SpacyNLP]

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """The component's default config (see parent class for full docstring)."""
        return {
            # Flag to check whether to split intents
            "intent_tokenization_flag": False,
            # Symbol on which intent should be split
            "intent_split_symbol": "_",
            # Regular expression to detect tokens
            "token_pattern": None,
            # Symbol on which prefix should be split
            "prefix_separator_symbol": None,
        }

    @staticmethod
    def required_packages() -> List[Text]:
        """Any extra python dependencies required for this component to run."""
        return ["spacy"]

    def _get_doc(self, message: Message, attribute: Text) -> Optional["Doc"]:
        return message.get(SPACY_DOCS[attribute])

    def tokenize(self, message: Message, attribute: Text) -> List[Token]:
        """Tokenizes the text of the provided attribute of the incoming message."""
        doc = self._get_doc(message, attribute)
        if not doc:
            return []

        tokens = [
            Token(
                t.text, t.idx, lemma=t.lemma_, data={POS_TAG_KEY: self._tag_of_token(t)}
            )
            for t in doc
            if t.text and t.text.strip()
        ]

        return self._apply_token_pattern(tokens)

    @staticmethod
    def _tag_of_token(token: Any) -> Text:
        if spacy.about.__version__ > "2" and token._.has("tag"):
            return token._.get("tag")
        else:
            return token.tag_



"""MCP tools for bot project file operations - lightweight implementation."""

from pathlib import Path
from typing import Dict, List, Optional

import structlog

from rasa.builder.copilot.mcp_server.models import (
    FileContentResponse,
    FileListResponse,
    FileUpdateFailure,
    ReadFilesResponse,
    UpdateFilesResponse,
    WriteFileResponse,
)
from rasa.builder.project_generator import (
    bot_file_paths,
    get_bot_files,
    is_restricted_path,
    unsafe_write_to_bot_files,
)
from rasa.utils.io import InvalidPathException

structlogger = structlog.get_logger()


async def list_files(project_folder: str) -> FileListResponse:
    """List all bot project files with directory structure.

    Args:
        project_folder: Path to the project folder

    Returns:
        FileListResponse containing files organized by directory with a tree structure.
    """
    try:
        project_path = Path(project_folder)

        # Collect all allowed files
        all_files = list(bot_file_paths(project_path))
        all_file_paths = [
            str(file_path.relative_to(project_path)) for file_path in all_files
        ]

        # Organize files by directory
        file_tree: Dict[str, List[str]] = {}
        root_files: List[str] = []

        for file_path in sorted(all_file_paths):
            if "/" in file_path:
                # File in a subdirectory
                parts = file_path.split("/")
                directory = parts[0]
                filename = "/".join(parts[1:])

                if directory not in file_tree:
                    file_tree[directory] = []
                file_tree[directory].append(filename)
            else:
                # Root level file
                root_files.append(file_path)

        # Build a human-readable tree structure
        tree_lines = []
        tree_lines.append("Project Structure:")
        tree_lines.append(".")

        # Add root files
        for f in root_files:
            tree_lines.append(f"├── {f}")

        # Add directories and their contents
        dirs = sorted(file_tree.keys())
        for i, directory in enumerate(dirs):
            is_last_dir = i == len(dirs) - 1
            prefix = "└──" if is_last_dir else "├──"
            tree_lines.append(f"{prefix} {directory}/")

            files_in_dir = sorted(file_tree[directory])
            for j, f in enumerate(files_in_dir):
                is_last_file = j == len(files_in_dir) - 1
                if is_last_dir:
                    file_prefix = "    └──" if is_last_file else "    ├──"
                else:
                    file_prefix = "│   └──" if is_last_file else "│   ├──"
                tree_lines.append(f"{file_prefix} {f}")

        tree_view = "\n".join(tree_lines)

        return FileListResponse(
            success=True,
            tree=tree_view,
            files=sorted(all_file_paths),
            count=len(all_file_paths),
            directories=list(file_tree.keys()),
        )

    except Exception as e:
        structlogger.error(
            "mcp_server.tools.file_operations.list_files.error",
            event_info="MCP tool failed to list files",
            error=str(e),
        )
        return FileListResponse(
            success=False,
            files=[],
            count=0,
            directories=[],
            error=f"Failed to list files: {e!s}",
        )


async def read_assistant_files(
    project_folder: str,
    exclude_docs: bool = True,
    allowed_extensions: Optional[str] = None,
) -> ReadFilesResponse:
    """Read all bot project files.

    Args:
        project_folder: Path to the project folder
        exclude_docs: Whether to exclude the docs directory (default: True)
        allowed_extensions: Comma-separated list of file extensions to include
                          (e.g., "yaml,yml,py"). If not provided, includes all files.

    Returns:
        ReadFilesResponse containing a mapping of file paths to their contents.
    """
    try:
        project_path = Path(project_folder)

        # Parse allowed extensions if provided
        extensions = None
        if allowed_extensions:
            extensions = [ext.strip().lower() for ext in allowed_extensions.split(",")]

        bot_files: Dict[str, Optional[str]] = get_bot_files(
            project_path, extensions, exclude_docs
        )

        return ReadFilesResponse(files=bot_files, count=len(bot_files))

    except Exception as e:
        structlogger.error(
            "mcp_server.tools.file_operations.read_assistant_files.error",
            event_info="MCP tool failed to read bot files",
            error=str(e),
        )
        return ReadFilesResponse(
            files={},
            count=0,
            error=f"Failed to read bot files: {e!s}",
        )


async def get_file_content(project_folder: str, file_path: str) -> FileContentResponse:
    """Get the content of a specific bot project file.

    Args:
        project_folder: Path to the project folder
        file_path: Relative path to the file within the project

    Returns:
        FileContentResponse containing the file content and metadata.

    Raises:
        InvalidPathException: If file path is invalid or access is denied
    """
    try:
        project_path = Path(project_folder)
        full_path = project_path / file_path

        # Security check
        if is_restricted_path(project_path, full_path):
            raise InvalidPathException(
                f"Access denied to '{file_path}'. Cannot access hidden files, "
                "__pycache__, or files outside the project directory."
            )

        if not full_path.exists():
            return FileContentResponse(
                file_path=file_path,
                content=None,
                exists=False,
                error=f"File not found: {file_path}",
            )

        content = full_path.read_text(encoding="utf-8")
        return FileContentResponse(
            file_path=file_path,
            content=content,
            exists=True,
        )

    except Exception as e:
        structlogger.error(
            "mcp_server.tools.file_operations.get_file_content.error",
            event_info="MCP tool failed to get file content",
            file_path=file_path,
            error=str(e),
        )
        return FileContentResponse(
            file_path=file_path,
            content=None,
            exists=False,
            error=f"Failed to get file content: {e!s}",
        )


async def write_file(
    project_folder: str, file_path: str, content: str
) -> WriteFileResponse:
    """Write content to a bot project file.

    This will create or overwrite the file at the specified path.

    Args:
        project_folder: Path to the project folder
        file_path: Relative path to the file within the project (e.g., "domain.yml")
        content: The content to write to the file

    Returns:
        WriteFileResponse indicating success or failure

    """
    try:
        project_path = Path(project_folder)
        unsafe_write_to_bot_files(
            project_path, {file_path: content}, fail_on_restricted_path=True
        )
        return WriteFileResponse(
            success=True,
            file_path=file_path,
            message=f"Successfully wrote {len(content)} characters to {file_path}",
        )

    except Exception as e:
        structlogger.error(
            "mcp_server.tools.file_operations.write_file.error",
            event_info="Failed to write file",
            file_path=file_path,
            error=str(e),
        )
        return WriteFileResponse(
            success=False,
            file_path=file_path,
            message=f"Failed to write file: {e!s}",
        )


async def update_files(
    project_folder: str, files: Dict[str, str]
) -> UpdateFilesResponse:
    """Update multiple bot project files at once.

    This is useful for making coordinated changes across multiple files.

    Args:
        project_folder: Path to the project folder
        files: Dictionary mapping file paths to their new contents

    Returns:
        UpdateFilesResponse with results for each file
    """
    try:
        updated = []
        failed = []

        for file_path, content in files.items():
            result = await write_file(project_folder, file_path, content)

            if result.success:
                updated.append(file_path)
            else:
                failed.append(
                    FileUpdateFailure(
                        file_path=file_path,
                        error=result.message,
                    )
                )

        success = len(failed) == 0
        message = f"Updated {len(updated)} file(s)"
        if failed:
            message += f", {len(failed)} failed"

        return UpdateFilesResponse(
            success=success,
            updated=updated,
            failed=failed,
            message=message,
        )

    except Exception as e:
        structlogger.error(
            "mcp_server.tools.file_operations.update_files.error",
            event_info="Failed to update files",
            error=str(e),
        )
        return UpdateFilesResponse(
            success=False,
            updated=[],
            failed=[FileUpdateFailure(error=str(e))],
            message=f"Failed to update files: {e!s}",
        )



from pathlib import Path
from typing import Dict, Generator, List, Optional

import structlog

from rasa.builder.exceptions import InvalidFileContentError
from rasa.builder.models import BotFiles
from rasa.shared.constants import (
    DEFAULT_MODELS_PATH,
)
from rasa.utils.io import InvalidPathException, subpath

DEFAULT_COMMIT_MESSAGE = "Update files"
structlogger = structlog.get_logger()


def get_bot_files(
    project_folder: Path,
    allowed_file_extensions: Optional[List[str]] = None,
    exclude_docs_directory: bool = False,
    exclude_models_directory: bool = True,
) -> BotFiles:
    """Get the current bot files by reading from disk.

    Args:
        project_folder: Path to the project folder
        allowed_file_extensions: Optional list of file extensions to include.
            If None, fetch all files. If provided, only fetch files with matching
            extensions. Use `""` empty string to allow files with no extensions.
        exclude_docs_directory: Optional boolean indicating whether to exclude.
        exclude_models_directory: Optional boolean indicating whether to exclude.

    Returns:
        Dictionary of file contents with relative paths as keys
    """
    bot_files: BotFiles = {}

    for file in bot_file_paths(project_folder, exclude_models_directory):
        relative_path = file.relative_to(project_folder)

        # Exclude the docs directory if specified
        if exclude_docs_directory and relative_path.parts[0] == "docs":
            continue

        # Exclude the files by file extensions if specified
        if allowed_file_extensions is not None:
            allowed_file_extensions = [ext.lower() for ext in allowed_file_extensions]
            if file.suffix.lstrip(".").lower() not in allowed_file_extensions:
                continue
        # Read file content and store with relative path as key
        try:
            bot_files[relative_path.as_posix()] = file.read_text(encoding="utf-8")
        except Exception as e:
            structlogger.debug(
                "project_generator.get_bot_files.error",
                error=str(e),
                file_path=file.as_posix(),
            )
            bot_files[relative_path.as_posix()] = None
    return bot_files


def bot_file_paths(
    project_folder: Path, exclude_models_directory: bool = True
) -> Generator[Path, None, None]:
    """Get the paths of all bot files.

    Args:
        project_folder: Path to the project folder
        exclude_models_directory: Optional boolean indicating whether to exclude.

    Returns:
        Generator of file paths
    """
    for file in project_folder.glob("**/*"):
        # Skip directories
        if not file.is_file() or is_restricted_path(
            project_folder, file, exclude_models_directory
        ):
            continue

        yield file


def is_restricted_path(
    project_folder: Path, path: Path, exclude_models_directory: bool = True
) -> bool:
    """Check if the path is restricted.

    These paths are excluded from deletion and editing by the user.

    Args:
        project_folder: Path to the project folder
        path: Path to the file or directory
        exclude_models_directory: Optional boolean indicating whether to exclude.

    Returns:
        True if the path is restricted, False otherwise
    """
    relative_path = path.relative_to(project_folder)

    # Skip hidden files and directories (any path component starting with '.')
    # as well as `__pycache__` folders
    if any(part.startswith(".") for part in relative_path.parts):
        return True

    if "__pycache__" in relative_path.parts:
        return True

    # exclude the project_folder / models folder if specified
    if exclude_models_directory and relative_path.parts[0] == DEFAULT_MODELS_PATH:
        return True

    return False


def unsafe_write_to_bot_files(
    project_folder: Path,
    files: Dict[str, Optional[str]],
    fail_on_restricted_path: bool = True,
) -> None:
    """Write content to bot project files.

    This does NOT acquire a git operation lock, make sure the caller
    has a lock on files to prevent concurrent writes.

    Args:
        project_folder: Path to the project folder
        files: Dictionary mapping file names to their content
        fail_on_restricted_path: Optional boolean indicating whether to fail
            on restricted paths.
    """
    for filename, content in files.items():
        file_path = path_relative_to_project(project_folder, filename)
        # Disallow updates inside .rasa project metadata directory
        if any(
            part.startswith(".") for part in file_path.relative_to(project_folder).parts
        ):
            # silently ignore hidden paths
            if fail_on_restricted_path:
                raise InvalidPathException(
                    f"This file or folder is restricted from editing: {file_path}"
                )
            continue
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_text(content or "", encoding="utf-8")


def path_relative_to_project(project_folder: Path, filename: str) -> Path:
    """Get the relative path of a file or directory to the project folder."""
    return Path(subpath(str(project_folder), filename))


def validate_file_content_transitions(
    new_files: Dict[str, Optional[str]],
    current_files: BotFiles,
    project_folder: Path,
) -> None:
    """Validate file content state transitions.

    Rules:
    - binary (null) -> string: REJECT - cannot modify binary content
    - binary (null) -> null: OK - preserve file unchanged
    - text (string) -> null: REJECT - use "" for empty file
    - text (string) -> string: OK - write new content
    - doesn't exist -> null: REJECT - cannot create binary via API
    - doesn't exist -> string: OK - create new file

    Args:
        new_files: Dictionary mapping file names to their new content
        current_files: Dictionary of current file contents
        project_folder: Path to the project folder for path normalization

    Raises:
        InvalidFileContentError: If a state transition is not allowed
    """
    abs_project_folder = project_folder.resolve()

    for filename, new_content in new_files.items():
        # Normalize the filename to match the format used in current_files
        # (relative POSIX paths like "image.png" instead of "./image.png")
        normalized_path = path_relative_to_project(abs_project_folder, filename)
        normalized_filename = normalized_path.relative_to(abs_project_folder).as_posix()

        current_content: str | None = current_files.get(normalized_filename)
        file_exists: bool = normalized_filename in current_files

        if new_content is None:
            # New content is null - only valid for existing binary files
            if not file_exists:
                raise InvalidFileContentError(
                    f"Cannot create binary file '{filename}' through this API. "
                )
            if current_content is not None:
                # Current file is text, new content is null
                raise InvalidFileContentError(
                    f"Cannot set text file '{filename}' to null. "
                    f'Use an empty string ("") if you want an empty file, '
                    f"or omit the file from the request to delete it."
                )
            # current_content is None (binary) and new_content is None -> OK
        else:
            # New content is string
            if file_exists and current_content is None:
                # Current file is binary, new content is text
                raise InvalidFileContentError(
                    f"Cannot modify binary file '{filename}' through this API. "
                    f"Binary files can only be preserved (send with content: null) "
                    f"or deleted (omit from request)."
                )
